(**************************************************************************)
(*                                                                        *)
(*  IMAP add-on for the Weasel mail server                                *)
(*  Copyright (C) 2017   Peter Moylan                                     *)
(*                                                                        *)
(*  This program is free software: you can redistribute it and/or modify  *)
(*  it under the terms of the GNU General Public License as published by  *)
(*  the Free Software Foundation, either version 3 of the License, or     *)
(*  (at your option) any later version.                                   *)
(*                                                                        *)
(*  This program is distributed in the hope that it will be useful,       *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU General Public License for more details.                          *)
(*                                                                        *)
(*  You should have received a copy of the GNU General Public License     *)
(*  along with this program.  If not, see <http://www.gnu.org/licenses/>. *)
(*                                                                        *)
(*  To contact author:   http://www.pmoylan.org   peter@pmoylan.org       *)
(*                                                                        *)
(**************************************************************************)

IMPLEMENTATION MODULE Envelopes;

        (********************************************************)
        (*                                                      *)
        (*       The ENVELOPE structure of a mail message       *)
        (*                                                      *)
        (*  Programmer:         P. Moylan                       *)
        (*  Started:            20 June 2017                    *)
        (*  Last edited:        5 July 2017                     *)
        (*  Status:             Seems to be working             *)
        (*                                                      *)
        (********************************************************)

(****************************************************************************)
(*                                                                          *)
(* The standard relevant to this module is RFC 2060                         *)
(*                                                                          *)
(* The envelope is a parenthesised list that returns some of the header     *)
(* lines.  This is meaningful mostly at the outermost level, but a          *)
(* BODYSTRUCTURE request has to include the envelope of any embedded        *)
(* message/rfc822 part, therefore we have to deal with it in this module.   *)
(*                                                                          *)
(****************************************************************************)


IMPORT Strings;

FROM Parser IMPORT
    (* proc *)  Match;

FROM Names IMPORT
    (* type *)  FilenameString;

<* IF TESTING THEN *>

FROM FileOps IMPORT
    (* const*)  NoSuchChannel,
    (* type *)  ChanId, FilePos,
    (* proc *)  OpenOldFile, CloseFile, ReadLine, CurrentPosition,
                SetPosition, FWriteString;

<* END *>

FROM BufferedTextStreams IMPORT
    (* type *)  TextStream,
    (* proc *)  TSReadLine;

FROM STextIO IMPORT
    (* proc *)  WriteString, WriteLn;

FROM MiscFuncs IMPORT
    (* proc *)  StringMatch, AppendCard, WriteCard;

FROM Conversions IMPORT
    (* proc *)  CardinalToStringLJ;

FROM Storage IMPORT
    (* proc *)  ALLOCATE, DEALLOCATE;

(************************************************************************)

CONST
    Nul = CHR(0);  Tab = CHR(9);  Space = ' ';  CtrlZ = CHR(26);

TYPE
    CharSet = SET OF CHAR;

(************************************************************************)
(*                         CHECK HEADER LABEL                           *)
(************************************************************************)

PROCEDURE HeaderMatch (VAR (*IN*) line: ARRAY OF CHAR;
                                  header: ARRAY OF CHAR): BOOLEAN;

    (* Returns TRUE iff header':' is a leading substring of line. *)

    VAR N: CARDINAL;
        test: ARRAY [0..1023] OF CHAR;

    BEGIN
        N := Strings.Length (header);

        (* Check for the colon first, because this will eliminate  *)
        (* most cases.                                             *)

        IF line[N] = ':' THEN
            Strings.Assign (line, test);
            test[N] := Nul;
            Strings.Capitalize (test);
            Strings.Capitalize (header);
            RETURN Strings.Equal (test, header);
        ELSE
            RETURN FALSE;
        END (*IF*);

    END HeaderMatch;

(************************************************************************)
(*                               ENVELOPE                               *)
(************************************************************************)

TYPE
    HeaderLine = ARRAY [0..1023] OF CHAR;

    EmailAddress = RECORD
                       PersonalName: HeaderLine;
                       AtDomainList: HeaderLine;
                       MailboxName:  HeaderLine;
                       HostName:     HeaderLine;
                   END (*RECORD*);

    EmailAddressList = POINTER TO
                           RECORD
                               next: EmailAddressList;
                               this: EmailAddress;
                           END (*RECORD*);

    EnvelopePtr = POINTER TO
                      RECORD
                          date:        HeaderLine;
                          subject:     HeaderLine;
                          from:        EmailAddressList;
                          sender:      EmailAddressList;
                          reply_to:    EmailAddressList;
                          to:          EmailAddressList;
                          cc:          EmailAddressList;
                          bcc:         EmailAddressList;
                          in_reply_to: HeaderLine;
                          message_id:  HeaderLine;
                      END (*RECORD*);

(************************************************************************)

PROCEDURE DiscardAddressList (VAR (*INOUT*) AL: EmailAddressList);

    (* Disposes of an address list. *)

    VAR p: EmailAddressList;

    BEGIN
        WHILE AL <> NIL DO
            p := AL^.next;
            DISPOSE (AL);
            AL := p;
        END (*WHILE*);
    END DiscardAddressList;

(************************************************************************)

PROCEDURE DiscardEnvelope (VAR (*INOUT*) env: EnvelopePtr);

    (* Disposes of an envelope structure. *)

    BEGIN
        IF env <> NIL THEN
            DiscardAddressList (env^.from);
            DiscardAddressList (env^.sender);
            DiscardAddressList (env^.reply_to);
            DiscardAddressList (env^.to);
            DiscardAddressList (env^.cc);
            DiscardAddressList (env^.bcc);
            DISPOSE (env);
        END (*IF*);
    END DiscardEnvelope;

(************************************************************************)
(*              PARSING THE HEADER TO EXTRACT ENVELOPE DATA             *)
(************************************************************************)

PROCEDURE ParseEnvelope (TS: TextStream): EnvelopePtr;

    (* Extracts the envelope date from the header lines. *)

    VAR line: HeaderLine;

    (********************************************************************)

    PROCEDURE DeleteLeading (N: CARDINAL);

        (* Deletes the leading N characters of line, plus any space or  *)
        (* tab characters that follow.                                  *)

        BEGIN
            WHILE (line[N] = Space) OR (line[N] = Tab) DO
                INC (N);
            END (*WHILE*);
            IF N > 0 THEN
                Strings.Delete (line, 0, N);
            END (*IF*);
        END DeleteLeading;

    (********************************************************************)

    PROCEDURE StripSpaces (VAR (*INOUT*) str: ARRAY OF CHAR);

        (* Removes leading and trailing spaces from str. *)

        VAR k: CARDINAL;

        BEGIN
            k := 0;
            WHILE str[k] = ' ' DO
                INC (k);
            END (*WHILE*);
            IF k > 0 THEN
                Strings.Delete (str, 0, k);
            END (*IF*);
            k := Strings.Length(str);
            WHILE (k > 0) AND (str[k-1] = ' ') DO
                DEC (k);
            END (*WHILE*);
            str[k] := Nul;
        END StripSpaces;

    (********************************************************************)

    PROCEDURE ParseAddress (VAR (*OUT*) addr: EmailAddress);

        (* Parses 'line' to extract the address components, using       *)
        (* RFC821 <path> syntax.  At least, that's the theory.  In      *)
        (* practice, two forms of the syntax occur to the exclusion of  *)
        (* all others:                                                  *)
        (*     personalname <user@domain>                               *)
        (*     user@domain (personalname)                               *)
        (* but I'll also allow for                                      *)
        (*     <user@domain> (personalname)                             *)
        (*     (personalname) <user@domain>                             *)
        (* and that's all I bother to check.                            *)

        VAR posat, posless, posparen, length: CARDINAL;
            foundat, foundless, foundparen: BOOLEAN;
            UAD: HeaderLine;

        BEGIN
            addr.AtDomainList[0] := Nul;
            addr.PersonalName := line;
            UAD := line;
            Strings.FindNext ('<', line, 0, foundless, posless);
            Strings.FindNext ('(', line, 0, foundparen, posparen);
            IF foundparen THEN
                IF foundless THEN
                    IF posparen < posless THEN
                        Strings.Delete (addr.PersonalName, 0, 1);
                        Strings.FindNext (')', addr.PersonalName, 0, foundparen, posparen);
                        IF NOT foundparen THEN
                            posparen := posless;
                        END (*IF*);
                        addr.PersonalName[posparen] := Nul;
                        Strings.Delete (UAD, 0, posless);
                    ELSE
                        UAD[posparen] := Nul;
                        Strings.Delete (addr.PersonalName, 0, posparen+1);
                        Strings.FindNext (')', addr.PersonalName, 0, foundparen, posparen);
                        IF foundparen THEN
                            addr.PersonalName[posparen] := Nul;
                        END (*IF*);
                    END (*IF*);
                ELSE
                    (* Found '(' but not '<' *)
                    UAD[posparen] := Nul;
                    Strings.Delete (addr.PersonalName, 0, posparen+1);
                END (*IF*);
            ELSIF foundless THEN
                (* Found '<' but not '(' *)
                addr.PersonalName[posless] := Nul;
                IF posless > 0 THEN
                    Strings.Delete (UAD, 0, posless);
                END (*IF*);
            ELSE
                (* Did not find '(' *)
                addr.PersonalName := "";
            END (*IF*);

            (* We've separated the line into UAD and addr.PersonalName. *)
            (* Now clean these up by stripping spaces and removing the  *)
            (* <> brackets, if present, from UAD.  Also remove the      *)
            (* quotation marks, if present, from addr.PersonalName.     *)

            StripSpaces (addr.PersonalName);
            IF addr.PersonalName[0] = '"' THEN
                Strings.Delete (addr.PersonalName, 0, 1);
                length := Strings.Length(addr.PersonalName);
                IF (length > 0) AND (addr.PersonalName[length-1] = '"') THEN
                    addr.PersonalName[length-1] := Nul;
                END (*IF*);
            END (*IF*);
            StripSpaces (UAD);
            IF UAD[0] = '<' THEN
                Strings.Delete (UAD, 0, 1);
                length := Strings.Length(UAD);
                IF (length > 0) AND (UAD[length-1] = '>') THEN
                    UAD[length-1] := Nul;
                END (*IF*);
            END (*IF*);

            (* Finally, split UAD into user@domain.     *)

            addr.MailboxName := UAD;
            addr.HostName := UAD;
            Strings.FindNext ('@', UAD, 0, foundat, posat);
            IF foundat THEN
                addr.MailboxName[posat] := Nul;
                Strings.Delete (addr.HostName, 0, posat+1);
            ELSE
                addr.HostName[0] := Nul;
            END (*IF*);

        END ParseAddress;

    (********************************************************************)

    PROCEDURE ParseAddressList (): EmailAddressList;

        (* Parses 'line' to extract a comma-separated list of addresses. *)

        VAR result, tail: EmailAddressList;

        BEGIN
            NEW (result);
            tail := result;
            result^.next := NIL;
            ParseAddress (result^.this);
            DeleteLeading (0);
            WHILE line[0] = ',' DO
                DeleteLeading (1);
                NEW (tail^.next);
                tail := tail^.next;
                tail^.next := NIL;
                ParseAddress (tail^.this);
            END (*WHILE*);
            RETURN result;
        END ParseAddressList;

    (********************************************************************)

    PROCEDURE CopyAddressList (arg: EmailAddressList): EmailAddressList;

        (* Produces a copy of arg. *)

        VAR result, tail: EmailAddressList;

        BEGIN
            IF arg = NIL THEN
                result := NIL;
            ELSE
                NEW (result);
                result^.this := arg^.this;
                tail := result;
                tail^.next := NIL;
                arg := arg^.next;
                WHILE arg <> NIL DO
                    NEW (tail^.next);
                    tail := tail^.next;
                    tail^.next := NIL;
                    tail^.this := arg^.this;
                    arg := arg^.next;
                END (*WHILE*);
            END (*IF*);
            RETURN result;
        END CopyAddressList;

    (********************************************************************)

    VAR InHeader: BOOLEAN;
        result: EnvelopePtr;

    BEGIN
        NEW (result);
        WITH result^ DO
            date[0] := Nul;
            subject[0] := Nul;
            result^.from := NIL;
            result^.sender := NIL;
            result^.reply_to := NIL;
            result^.to := NIL;
            result^.cc := NIL;
            result^.bcc := NIL;
            in_reply_to[0] := Nul;
            message_id[0] := Nul;
        END (*WITH*);

        InHeader := TRUE;
        WHILE InHeader DO
            TSReadLine (TS, line);
            IF (line[0] = CtrlZ) OR (line[0] = Nul) THEN
                InHeader := FALSE;
            ELSIF HeaderMatch (line, "Date") THEN
                DeleteLeading (5);
                Strings.Assign (line, result^.date);
            ELSIF HeaderMatch (line, "From") THEN
                DeleteLeading (5);
                result^.from := ParseAddressList();
            ELSIF HeaderMatch (line, "Sender") THEN
                DeleteLeading (7);
                result^.sender := ParseAddressList();
            ELSIF HeaderMatch (line, "Reply-To") THEN
                DeleteLeading (9);
                result^.reply_to := ParseAddressList();
            ELSIF HeaderMatch (line, "To") THEN
                DeleteLeading (3);
                result^.to := ParseAddressList();
            ELSIF HeaderMatch (line, "Cc") THEN
                DeleteLeading (3);
                result^.cc := ParseAddressList();
            ELSIF HeaderMatch (line, "Bcc") THEN
                DeleteLeading (4);
                result^.bcc := ParseAddressList();
            ELSIF HeaderMatch (line, "Subject") THEN
                DeleteLeading (8);
                Strings.Assign (line, result^.subject);
            ELSIF HeaderMatch (line, "In-Reply-To") THEN
                DeleteLeading (12);
                Strings.Assign (line, result^.in_reply_to);
            ELSIF HeaderMatch (line, "Message-ID") THEN
                DeleteLeading (11);
                Strings.Assign (line, result^.message_id);
            END (*IF*);
        END (*WHILE*);

        (* The standard says that the reply_to and sender       *)
        (* fields should default to the from address if they    *)
        (* are missing, and that the IMAP server has the        *)
        (* responsibility to set this default.                  *)

        IF result^.reply_to = NIL THEN
            result^.reply_to := CopyAddressList (result^.from);
        END (*IF*);
        IF result^.sender = NIL THEN
            result^.sender := CopyAddressList (result^.from);
        END (*IF*);

        RETURN result;

    END ParseEnvelope;

(************************************************************************)
(*                  RETURNING THE ENVELOPE INFORMATION                  *)
(************************************************************************)

PROCEDURE AppendEnvelope (EnvData: EnvelopePtr;
                               VAR (*INOUT*) buffer: ARRAY OF CHAR);

    (* Appends the ENVELOPE information to buffer.  *)
    (* We do not destroy the EnvData structure.     *)

    (********************************************************************)

    PROCEDURE PutField (VAR (*IN*) value: ARRAY OF CHAR;  NILoption: BOOLEAN);

        BEGIN
            IF NILoption AND (value[0] = Nul) THEN
                Strings.Append ('NIL', buffer);
            ELSE
                Strings.Append ('"', buffer);
                Strings.Append (value, buffer);
                Strings.Append ('"', buffer);
            END (*IF*);
        END PutField;

    (********************************************************************)

    PROCEDURE PutAddr (VAR (*IN*) addr: EmailAddress);

        BEGIN
            Strings.Append ('(', buffer);
            PutField (addr.PersonalName, TRUE);
            Strings.Append (' ', buffer);
            PutField (addr.AtDomainList, TRUE);
            Strings.Append (' ', buffer);
            PutField (addr.MailboxName, TRUE);
            Strings.Append (' ', buffer);
            PutField (addr.HostName, TRUE);
            Strings.Append (')', buffer);
        END PutAddr;

    (********************************************************************)

    PROCEDURE PutAddrList (list: EmailAddressList);

        BEGIN
            Strings.Append ('(', buffer);
            IF list = NIL THEN
                Strings.Append ('(NIL NIL NIL NIL)', buffer);
            ELSE
                WHILE list <> NIL DO
                   PutAddr (list^.this);
                   list := list^.next;
                   IF list <> NIL THEN
                       Strings.Append (' ', buffer);
                   END (*IF*);
                END (*WHILE*);
            END (*IF*);
            Strings.Append (')', buffer);
        END PutAddrList;

    (********************************************************************)

    BEGIN
        IF EnvData = NIL THEN
            Strings.Append ('NIL', buffer);
        ELSE
            Strings.Append ('(', buffer);
            PutField (EnvData^.date, TRUE);
            Strings.Append (' ', buffer);
            PutField (EnvData^.subject, TRUE);
            Strings.Append (' ', buffer);
            PutAddrList (EnvData^.from);
            Strings.Append (' ', buffer);
            PutAddrList (EnvData^.sender);
            Strings.Append (' ', buffer);
            PutAddrList (EnvData^.reply_to);
            Strings.Append (' ', buffer);
            PutAddrList (EnvData^.to);
            Strings.Append (' ', buffer);
            PutAddrList (EnvData^.cc);
            Strings.Append (' ', buffer);
            PutAddrList (EnvData^.bcc);
            Strings.Append (' ', buffer);
            PutField (EnvData^.in_reply_to, TRUE);
            Strings.Append (' ', buffer);
            PutField (EnvData^.message_id, TRUE);
            Strings.Append (')', buffer);
            (*DISPOSE (EnvData);*)
        END (*IF*);
    END AppendEnvelope;

(************************************************************************)

<* IF TESTING THEN *>

PROCEDURE DumpEnvelope (cid: ChanId;  env: EnvelopePtr);

    (* For use in debugging.  Writes the envelope to a file. *)

    VAR buffer: ARRAY [0..4095] OF CHAR;

    BEGIN
        buffer[0] := Nul;
        AppendEnvelope (env, buffer);
        FWriteString (cid, buffer);
    END DumpEnvelope;

(************************************************************************)

PROCEDURE DebugDumpEnvelope (message: ARRAY OF CHAR;  env: EnvelopePtr);

    (* For use in debugging.  Writes the envelope to a standard output. *)

    VAR buffer: ARRAY [0..4095] OF CHAR;

    BEGIN
        WriteString (message);  WriteLn;
        buffer[0] := Nul;
        AppendEnvelope (env, buffer);
        WriteString (buffer);  WriteLn;
    END DebugDumpEnvelope;

<* END *>

(************************************************************************)

(*
PROCEDURE ReturnEnvelope (VAR (*IN*) file: FilenameString;
                               VAR (*INOUT*) buffer: ARRAY OF CHAR);

    (* Appends the ENVELOPE information to buffer.  This is for the     *)
    (* outermost envelope, and only for the case where the file is not  *)
    (* already open, so we might not need it.                           *)

    VAR EnvData: EnvelopePtr;
        cid: ChanId;

    BEGIN
        cid := OpenOldFile (file, FALSE);
        EnvData := ParseEnvelope(cid);
        CloseFile (cid);
        AppendEnvelope (EnvData, buffer);
    END ReturnEnvelope;
*)

(************************************************************************)

END Envelopes.

