(**************************************************************************)
(*                                                                        *)
(*  IMAP add-on for the Weasel mail server                                *)
(*  Copyright (C) 2017   Peter Moylan                                     *)
(*                                                                        *)
(*  This program is free software: you can redistribute it and/or modify  *)
(*  it under the terms of the GNU General Public License as published by  *)
(*  the Free Software Foundation, either version 3 of the License, or     *)
(*  (at your option) any later version.                                   *)
(*                                                                        *)
(*  This program is distributed in the hope that it will be useful,       *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU General Public License for more details.                          *)
(*                                                                        *)
(*  You should have received a copy of the GNU General Public License     *)
(*  along with this program.  If not, see <http://www.gnu.org/licenses/>. *)
(*                                                                        *)
(*  To contact author:   http://www.pmoylan.org   peter@pmoylan.org       *)
(*                                                                        *)
(**************************************************************************)

IMPLEMENTATION MODULE MailStructure;

        (********************************************************)
        (*                                                      *)
        (*         The MIME structure of a mail message         *)
        (*                                                      *)
        (*  Programmer:         P. Moylan                       *)
        (*  Started:            1 August 2004                   *)
        (*  Last edited:        12 July 2017                    *)
        (*  Status:             Working                         *)
        (*                                                      *)
        (********************************************************)

(****************************************************************************)
(*                                                                          *)
(*                            GENERAL COMMENTS                              *)
(*                                                                          *)
(* The standards relevant to this module are:                               *)
(*                                                                          *)
(*      RFC 2045    MIME and MIME header fields                             *)
(*      RFC 2060    IMAP                                                    *)
(*      RFC 2046    MIME types                                              *)
(*                                                                          *)
(* The possible MIME parts can be broken into three groups:                 *)
(*  1.  Simple parts, i.e. parts without further substructure               *)
(*  2.  multipart, consisting of several concatenated parts                 *)
(*  3.  message: an embedded message, which has a header and a body, where  *)
(*          the body might itself have MIME components. A top-level message *)
(*          can probably be considered to be in this category.              *)
(* RFC 2045 does not allow for any other possibility, although future       *)
(* standards could change this.  I have not yet decided whether I should    *)
(* formally create three separate data types for these three cases, or aim  *)
(* at a representation that handles the union of the three.  At present I   *)
(* have a single data type, which means carrying some redundant information.*)
(*                                                                          *)
(* RFC 2046 is largely irrelevant, but it does define the "multipart" and   *)
(* "message" types. The multipart type covers the preamble and epilogue     *)
(* (which I have allowed for). For the message/rfc822 subtype, we must      *)
(* recognise that this type has no terminator other than running out of     *)
(* data (section 5.1.2). To handle this we use a "terminator stack" that    *)
(* includes the terminators of enclosing parts.  It is not yet clear how    *)
(* well this compensates for ill-formed messages, but tests show that it    *)
(* corrects for some missing boundary codes.                                *)
(*                                                                          *)
(* The type message/partial is also defined, but I have not yet decided     *)
(* whether this is relevant to us. The message/external-body can, I think,  *)
(* be ignored because the explanation is so incoherent that nobody is       *)
(* going to want to implement this case.                                    *)
(*                                                                          *)
(* IMAP queries can request the following details (among others that are    *)
(* less relevant here) about a message:                                     *)
(*                                                                          *)
(*  ENVELOPE                                                                *)
(*      A parenthesised list that returns some of the header lines.  This   *)
(*      is relevant only for the overall message, or for an embedded        *)
(*      message/rfc822 component.  For the sake of readability I have       *)
(*      moved all envelope operations to a separate module Envelopes.       *)
(*                                                                          *)
(*  BODYSTRUCTURE                                                           *)
(*      A parenthesised list like                                           *)
(*        ("TEXT" "PLAIN" ("CHARSET" "US-ASCII") NIL NIL "7BIT" 2279 48)    *)
(*      that gives some very basic information like content type and in     *)
(*      some cases size in bytes and lines.  This example is for a simple   *)
(*      text message.  For a multipart list, the result includes sublists.  *)
(*                                                                          *)
(*  PARTS                                                                   *)
(*      This is what is fetched by the FETCH BODY[section]<partial>         *)
(*      command. We return actual message content of all or a specified     *)
(*      part of a message.                                                  *)
(*                                                                          *)
(* I have not yet decided whether a single data structure should allow for  *)
(* all of these, or whether separate parses should be used.  My present     *)
(* implementation in Messages.mod uses a common BuildStructureData as the   *)
(* preparation for both BODYSTRUCTURE and PARTS.                            *)
(*                                                                          *)
(****************************************************************************)


<* IF TESTING THEN *>
    IMPORT FileOps;
    IMPORT FileSys;
    FROM Types IMPORT CARD64;
    FROM Envelopes IMPORT DumpEnvelope, DebugDumpEnvelope;
<* END *>

IMPORT Strings;

FROM Types IMPORT
    (* type *)  INT64;

FROM LONGLONG IMPORT
    (* proc *)  Diff64;

FROM STextIO IMPORT
    (* proc *)  WriteString, WriteLn;

FROM Conversions IMPORT
    (* proc *)  CardinalToStringLJ;

FROM Storage IMPORT
    (* proc *)  ALLOCATE, DEALLOCATE;

FROM FileOps IMPORT
    (* type *)  FilePos;

FROM BufferedTextStreams IMPORT
    (* type *)  TextStream,
    (* proc *)  OpenForReading, TSFileOpened, CloseTS, TSReadLine,
                TSStartPosition, TSCurrentPosition, TSSetPosition, TSBackup;

FROM Replies IMPORT
    (* type *)  ReplyCxt,
    (* proc *)  PartialReply;

FROM Envelopes IMPORT
    (* type *)  EnvelopePtr,
    (* proc *)  HeaderMatch, ParseEnvelope, AppendEnvelope,
                DiscardEnvelope;

FROM Parser IMPORT
    (* proc *)  Match;

FROM MiscFuncs IMPORT
    (* proc *)  StringMatch, AppendCard, WriteCard;

FROM Names IMPORT
    (* type *)  FilenameString;

(************************************************************************)

CONST
    Nul = CHR(0);  Tab = CHR(9);  Space = ' ';  CtrlZ = CHR(26);

TYPE
    CharSet = SET OF CHAR;

CONST
    Digits = CharSet{'0'..'9'};

(************************************************************************)
(*                                                                      *)
(*                           STRUCTURE DATA                             *)
(*                                                                      *)
(* A StructureData tree represents the MIME structure of a message.     *)
(* By walking through this tree we can produce a BODYSTRUCTURE response *)
(* that summarises the structure.  More importantly, we can use this    *)
(* data structure to work out the location and size of a component of   *)
(* the message specified by a FETCH parameter.                          *)
(*                                                                      *)
(************************************************************************)

TYPE
    String32 = ARRAY [0..31] OF CHAR;

    ParamPtr = POINTER TO
                   RECORD
                       next: ParamPtr;
                       attribute, value: ARRAY [0..63] OF CHAR;
                   END (*RECORD*);

    (* The BodyLineCount is needed only for bodies of type TEXT, but    *)
    (* the code is simpler if we gather it for all bodies.              *)
    (* The TotalLineCount is not always needed, but it must be correct  *)
    (* for all components because it will be needed if we are inside    *)
    (* a body of type MESSAGE/RFC822.                                   *)
    (* The bodysize must be correct for all non-multipart bodies.       *)
    (* The bodysize and totalsize are potentially needed for            *)
    (* any part, depending on what the client wants to fetch.           *)

    (* The StructureData describes a message or section of message      *)
    (* consisting of a header and a body.  (For the purpose of these    *)
    (* calculations, the blank line before the body is counted as part  *)
    (* of the header.)  For a nested component this header is a MIME    *)
    (* header, and otherwise it is a normal message header.  We         *)
    (* distinguish these cases by the Boolean flags in the record.      *)
    (* Nested components of a multipart message have their own data     *)
    (* recorded at the next level down of the tree, in "parts".         *)

    (* A MESSAGE/RFC822 nested component is a special case.  It has a   *)
    (* MIME header and also a normal message header.  To deal with this *)
    (* we store the information in two tree levels.  The first level    *)
    (* has the MIME header (but also has byte counts that are the sums  *)
    (* for the whole message).  The next level down has the actual      *)
    (* message header and message body.  We just have to remember that  *)
    (* those two levels are a single level as far as section numbering  *)
    (* is concerned.  It's possible, of course, that the body is        *)
    (* multipart, in which case we will have at least one further level *)
    (* as is normal for a multipart message.                            *)

    (* Logically, we should have two separate data types, one for the   *)
    (* overall message or an embedded MESSAGE/RFC822 message, and one   *)
    (* for one part of a multipart message.  So far I have been getting *)
    (* by with one record type to cover both cases, but this does       *)
    (* require the trick of recording a MESSAGE/RFC822 part in two tree *)
    (* levels, so that we can keep both the MIME headers and the        *)
    (* message headers.  The higher level holds the MIME headers and    *)
    (* has the is822part flag set.  The lower part has the ismessage    *)
    (* flag set, and contains the actual header and body of the message.*)
    (* I have not yet decided whether we could get a more elegant       *)
    (* solution by separating the two cases.                            *)

    StructurePtr = StructureData;
    StructureData = POINTER TO
                        RECORD
                            next: StructurePtr;
                            totalsize, TotalLineCount: CARDINAL;
                            headerstart, bodystart: FilePos;
                            headersize, bodysize: CARDINAL;
                            BodyLineCount: CARDINAL;
                            ismessage, multipart, is822part: BOOLEAN;
                            NoMore: BOOLEAN;
                            parts: StructurePtr;
                            envelope: EnvelopePtr;
                            firstparam: ParamPtr;

                            Content_type,
                            Content_subtype,
                            Content_ID,
                            Content_description,
                            Content_transfer_encoding: String32;

                        END (*RECORD*);

    (* The possible cases are a lot more restricted than is implied     *)
    (* by the above declarations.  We have the following cases.         *)
    (*                                                                  *)
    (* MESSAGE/RFC822 (which includes the case of the total message).   *)
    (*    starts with header lines (always), then a blank line, then    *)
    (*    the body.  The body may have substructure, depending on what  *)
    (*    is specified in the Content-Type header line.  Except at the  *)
    (*    top level this is of course a multipart component, therefore  *)
    (*    it is preceded by a MIME header.                              *)
    (*                                                                  *)
    (* multipart body                                                   *)
    (*    a sequence of components with no header.  All of the          *)
    (*    structure is in the parts; the only text that belongs to the  *)
    (*    body but not the parts are the separators, and the short      *)
    (*    filler that can come between the message header and the first *)
    (*    boundary code, and after the last part.                       *)
    (*                                                                  *)
    (* multipart component                                              *)
    (*    starts with MIME header lines (which are counted as header    *)
    (*    lines in the StructureData record), then a blank line, then   *)
    (*    the body.  The structure of the body is specified by the MIME *)
    (*    headers.                                                      *)
    (*                                                                  *)
    (* simple body                                                      *)
    (*    plain data with no substructure.                              *)

(************************************************************************)

PROCEDURE DiscardStructureParams (VAR (*INOUT*) q: ParamPtr);

    (* Disposes of the data structure. *)

    VAR qnext: ParamPtr;

    BEGIN
        WHILE q <> NIL DO
            qnext := q^.next;
            DISPOSE (q);
            q := qnext;
        END (*WHILE*);
    END DiscardStructureParams;

(************************************************************************)

PROCEDURE DiscardStructureData (VAR (*INOUT*) SD: StructureData);

    (* Disposes of the data structure. *)

    VAR p, next: StructureData;

    BEGIN
        IF SD <> NIL THEN
            DiscardStructureParams (SD^.firstparam);
            DiscardEnvelope (SD^.envelope);
            p := SD^.parts;
            WHILE p <> NIL DO
                next := p^.next;
                DiscardStructureData (p);
                p := next;
            END (*WHILE*);
            DISPOSE (SD);
        END (*IF*);
    END DiscardStructureData;

(************************************************************************)
(*                      DUMPING THE STRUCTURE DATA                      *)
(*                 needed only when testing this module                 *)
(************************************************************************)

<* IF TESTING THEN *>

VAR pos: CARDINAL;

PROCEDURE DumpParams (cid: FileOps.ChanId;  p: ParamPtr);

    BEGIN
        FileOps.FWriteChar (cid, '(');
        INC (pos);
        WHILE p <> NIL DO
            FileOps.FWriteString (cid, p^.attribute);
            FileOps.FWriteChar (cid, ' ');
            INC (pos, LENGTH(p^.attribute)+1);
            FileOps.FWriteString (cid, p^.value);
            INC (pos, LENGTH(p^.value));
            p := p^.next;
            IF p <> NIL THEN
                FileOps.FWriteChar (cid, ' ');
                INC (pos);
            END (*IF*);
        END (*WHILE*);
        FileOps.FWriteChar (cid, ')');
        INC (pos);
    END DumpParams;

(************************************************************************)

PROCEDURE WrLn (cid: FileOps.ChanId);

    BEGIN
        FileOps.FWriteLn (cid);
        pos := 0;
    END WrLn;

(************************************************************************)

PROCEDURE WrHex (cid: FileOps.ChanId;  val, ndigits: CARDINAL);

    BEGIN
        IF ndigits > 1 THEN
            WrHex (cid, val DIV 16, ndigits-1);
            val := val MOD 16;
        END (*IF*);
        IF val < 10 THEN
            FileOps.FWriteChar (cid, CHR(ORD('0') + val));
            INC (pos);
        ELSE
            FileOps.FWriteChar (cid, CHR(ORD('A') + val - 10));
            INC (pos);
        END (*IF*);
    END WrHex;

(************************************************************************)

PROCEDURE DumpHex64 (cid: FileOps.ChanId;  val: CARD64);

    BEGIN
        WrHex (cid, val.high, 4);  WrHex (cid, val.low, 4);
    END DumpHex64;

(********************************************************************)

PROCEDURE WrCard (cid: FileOps.ChanId;  val: CARDINAL);

    BEGIN
        IF val > 9 THEN
            WrCard (cid, val DIV 10);
            val := val MOD 10;
        END (*IF*);
        FileOps.FWriteChar (cid, CHR(ORD('0')+val));
        INC (pos);
    END WrCard;

(************************************************************************)

PROCEDURE WrChar (cid: FileOps.ChanId;  val: CHAR);

    BEGIN
        FileOps.FWriteChar (cid, val);
        INC (pos);
    END WrChar;

(************************************************************************)

PROCEDURE WrStr (cid: FileOps.ChanId;  val: ARRAY OF CHAR);

    BEGIN
        FileOps.FWriteString (cid, val);
        INC (pos, LENGTH(val));
    END WrStr;

(************************************************************************)

PROCEDURE Indent (cid: FileOps.ChanId;  level, adjust: CARDINAL);

    VAR j: CARDINAL;

    BEGIN
        FOR j := 1 TO 4*level-adjust DO
            FileOps.FWriteChar (cid, ' ');
            INC (pos);
        END (*FOR*);
    END Indent;

(************************************************************************)

PROCEDURE WrLabel0 (cid: FileOps.ChanId;  label: ARRAY OF CHAR);

    (* Like WrLabel, but with no extra indentation. *)

    BEGIN
        WrStr (cid, label);  WrStr (cid, ": ");
    END WrLabel0;

(************************************************************************)

PROCEDURE WrLabel (cid: FileOps.ChanId;  label: ARRAY OF CHAR;  level: CARDINAL);

    (* Writes the label at the beginning of a line, but indented. *)

    BEGIN
        Indent(cid, level, 0);
        WrLabel0 (cid, label);
    END WrLabel;

(************************************************************************)

PROCEDURE TabTo (cid: FileOps.ChanId;  level, col: CARDINAL);

    (* Move to column col + 4*level). *)

    VAR target: CARDINAL;

    BEGIN
        target := col + 4*level;
        WHILE pos < target DO
            WrChar (cid, ' ');
        END (*WHILE*);
    END TabTo;

(************************************************************************)

PROCEDURE DumpSD (cid: FileOps.ChanId;  SD: StructureData;  level: CARDINAL;
                  partnumprefix: ARRAY OF CHAR;  partnum: CARDINAL);

    (* Writes result to a file that the caller has already opened. *)

    VAR partnumlabel: ARRAY [0..63] OF CHAR;
        p: StructurePtr;  needcomma: BOOLEAN;

    BEGIN
        pos := 0;
        Strings.Assign (partnumprefix, partnumlabel);
        IF partnum > 0 THEN
            IF partnumprefix[0] <> Nul THEN
                Strings.Append (".", partnumlabel);
            END (*IF*);
            AppendCard (partnum, partnumlabel);
            Indent(cid, level, 2);
            FileOps.FWriteString (cid, "PART ");
            FileOps.FWriteString (cid, partnumlabel);
            WrLn (cid);
        END (*IF*);

        (* Headers specific to MIME. *)

        WrLabel (cid, "Content_type", level);
        WrStr (cid, SD^.Content_type);  WrChar(cid, '/');
        WrStr (cid, SD^.Content_subtype);
        TabTo (cid, level, 40);
        WrLabel0 (cid, "Content_ID");
        WrStr (cid, SD^.Content_ID);  WrLn(cid);
        WrLabel (cid, "Content_description", level);
        WrStr (cid, SD^.Content_description);
        TabTo (cid, level, 40);
        WrLabel0 (cid, "Content_transfer_encoding");
        WrStr (cid, SD^.Content_transfer_encoding);  WrLn(cid);

        (* Byte and line counts. *)

        WrLabel (cid, "TotalLineCount", level);
        WrCard (cid, SD^.TotalLineCount);
        TabTo (cid, level, 40);
        WrLabel0 (cid, "BodyLineCount");
        WrCard (cid, SD^.BodyLineCount);
        WrLn(cid);

        (* The four Boolean flags.  I could probably express this more  *)
        (* compactly, but why worry?                                    *)

        needcomma := FALSE;  Indent(cid, level, 0);
        WrChar (cid, '(');
        IF SD^.ismessage THEN
            WrStr (cid, "ismessage");
            needcomma := TRUE;
        END (*IF*);
        IF SD^.multipart THEN
            IF needcomma THEN
                WrStr (cid, ", ");
            END (*IF*);
            WrStr (cid, "multipart");
            needcomma := TRUE;
        END (*IF*);
        IF SD^.is822part THEN
            IF needcomma THEN
                WrStr (cid, ", ");
            END (*IF*);
            WrStr (cid, "is822part");
            needcomma := TRUE;
        END (*IF*);
        IF SD^.NoMore THEN
            IF needcomma THEN
                WrStr (cid, ", ");
            END (*IF*);
            WrStr (cid, "NoMore");
        END (*IF*);
        WrChar (cid, ')');

        (* Total size. *)

        TabTo (cid, level, 40);
        WrLabel0 (cid, "totalsize");
        WrCard (cid, SD^.totalsize);
        WrLn(cid);

        (* The numbers that let us extract body parts: header and   *)
        (* body start positions and sizes.                          *)

        WrLabel (cid, "headerstart", level);
        DumpHex64 (cid, SD^.headerstart);
        TabTo (cid, level, 40);
        WrLabel0 (cid, "headersize");
        WrCard (cid, SD^.headersize);
        WrLn(cid);
        WrLabel (cid, "bodystart", level);
        WrStr (cid, "  ");
        DumpHex64 (cid, SD^.bodystart);
        TabTo (cid, level, 40);
        WrLabel0 (cid, "bodysize");
        WrCard (cid, SD^.bodysize);
        WrLn(cid);

        (* Envelope and parameters. *)

        WrLabel (cid, "envelope", level);
        DumpEnvelope (cid, SD^.envelope);  WrLn(cid);
        WrLabel (cid, "params", level);
        DumpParams (cid, SD^.firstparam);  WrLn(cid);

        (* Now the nested parts. *)

        p := SD^.parts;
        IF SD^.is822part THEN
            Indent (cid, level, 0);
            WrStr(cid, "<The rfc-822 message>");  WrLn(cid);
            DumpSD (cid, p, level, partnumlabel, 0);
        ELSE
            partnum := 1;  INC(level);
            WHILE p <> NIL DO
                DumpSD (cid, p, level, partnumlabel, partnum);
                p := p^.next;
                INC (partnum);
            END (*WHILE*);
        END (*IF*);

    END DumpSD;

(************************************************************************)

PROCEDURE DumpStructureData (SD: StructureData);

    (* Writes result to file SDout.txt. *)

    CONST fname = "SDout.txt";

    VAR cid: FileOps.ChanId;
        dummy: BOOLEAN;

    BEGIN
        IF FileSys.Exists(fname) THEN
            FileSys.Remove(fname, dummy);
        END (*IF*);
        cid := FileOps.OpenNewFile (fname, FALSE);
        DumpSD (cid, SD, 0, "", 0);
        FileOps.CloseFile (cid);
    END DumpStructureData;

<* END *>

(************************************************************************)
(*                                                                      *)
(*                          BOUNDARY CHECKING                           *)
(*                                                                      *)
(* A multipart structure has explicit boundaries between its parts.     *)
(* A message/rfc822 component does not, so to find its end we have to   *)
(* hit the boundary of the enclosing structure, and then backtrack.     *)
(* An additional complication is that some messages don't seem to       *)
(* comply with the MIME standards, so that boundaries might be omitted. *)
(* To get around this, we keep a stack of boundary strings.  Mostly     *)
(* only the top of the stack is relevant, but the stack allows us to    *)
(* detect boundaries of the enclosing component of the message.         *)
(*                                                                      *)
(************************************************************************)

TYPE
    BoundaryString = ARRAY [0..73] OF CHAR;

    BoundaryStack = POINTER TO
                        RECORD
                            next: BoundaryStack;
                            this: BoundaryString;
                        END (*RECORD*);

(************************************************************************)

PROCEDURE AtThisBoundary (VAR (*IN*) line, boundary: ARRAY OF CHAR;
                            VAR (*OUT*) boundarysize: CARDINAL): CARDINAL;

    (* Checks line to see whether we are at end of file or end of       *)
    (* section.  The result codes are                                   *)
    (*     0    not at end of section                                   *)
    (*     1    line matches boundary, i.e. end of section              *)
    (*     2    line matches boundary'--'                               *)
    (* The possibility of end-of-file is handled by the caller.         *)
    (* The value of boundarysize is set to match the size of the        *)
    (* boundary code, if present, including the delimiting line         *)
    (* terminators which are officially part of the boundary code.  If  *)
    (* we are not at a boundary then boundarysize is returned as 0.     *)

    VAR code, L, M: CARDINAL;
        copy: ARRAY [0..2047] OF CHAR;

    BEGIN
        boundarysize := 0;
        code := 0;
        L := Strings.Length (line);
        M := Strings.Length (boundary);
        IF M = 0 THEN
            code := 0;
        ELSIF L = M THEN

            IF Strings.Equal (line, boundary) THEN
                code := 1;
                boundarysize := L+4;

                (* Technically, the <CRLF> before the boundary      *)
                (* code is part of the boundary, as is the <CRLF>   *)
                (* that terminates the boundary code.  That is why  *)
                (* we add 4 to the size.                            *)

            END (*IF*);

        ELSIF (L = M+2) AND (line[M] = '-') AND (line[M+1] = '-') THEN

            Strings.Assign (line, copy);
            copy[M] := Nul;
            IF Strings.Equal (copy, boundary) THEN
                code := 2;
                boundarysize := L+4;
            END (*IF*);

        END (*IF*);
        RETURN code;
    END AtThisBoundary;

(************************************************************************)

PROCEDURE AtBoundary (VAR (*IN*) line: ARRAY OF CHAR;
                      VAR (*INOUT*) stack: BoundaryStack;
                      VAR (*OUT*) boundarysize: CARDINAL): CARDINAL;

    (* Checks line to see whether we are at end of file or end of       *)
    (* section.  The result codes are                                   *)
    (*     0    not at end of section                                   *)
    (*     1    line matches boundary, i.e. end of section              *)
    (*     2    (a) line matches boundary'--'                           *)
    (*          (b) line matches a boundary deeper in the stack         *)
    (* In case 2 we pop the stack.  The value of  boundarysize is also  *)
    (* set to match the size of the boundary code, if present,          *)
    (* including the delimiting line terminators which are officially   *)
    (* part of the boundary code.  If we are not at a boundary, then    *)
    (* boundarysize is returned as 0.                                   *)

    VAR code: CARDINAL;
        current, next: BoundaryStack;
        popneeded: BOOLEAN;

    BEGIN
        boundarysize := 0;

        (* Start with a quick check of obvious cases. *)

        IF line[0] = CtrlZ THEN
            RETURN 2;
        ELSIF (line[0] <> '-') OR (line[1] <> '-') THEN
            RETURN 0;
        END (*IF*);

        code := AtThisBoundary (line, stack^.this, boundarysize);
        popneeded := (code <> 1) AND (stack^.next <> NIL);

        IF code = 0 THEN

            (* Look deeper into the stack, but never go deeper than *)
            (* the end-of-file boundary code.                       *)

            current := stack^.next;
            LOOP
                code := AtThisBoundary (line, current^.this, boundarysize);
                IF (code > 0) OR (current^.next = NIL) THEN
                    EXIT (*LOOP*);
                END (*IF*);
                current := current^.next;
            END (*WHILE*);

            popneeded := code <> 0;

        END (*IF*);

        (* Pop the stack if required. *)

        IF popneeded THEN
            next := stack^.next;
            DISPOSE (stack);
            stack := next;
        END (*IF*);

        RETURN code;

    END AtBoundary;

(************************************************************************)
(*                    WORKING OUT THE STRUCTURE DATA                    *)
(************************************************************************)

PROCEDURE ParseToken (VAR (*INOUT*) line: ARRAY OF CHAR;
                      VAR (*OUT*) token: ARRAY OF CHAR);

    CONST tspecials =  CharSet{Nul, "(", ")", "<", ">", "@",
                               ",", ";", ":", "\", '"',
                               "/", "[", "]", "?", "="};

    VAR j: CARDINAL;

    BEGIN
        j := 0;
        WHILE (j <= HIGH(line)) AND (j <= HIGH(token))
                   AND NOT (line[j] IN tspecials) DO
            token[j] := line[j];
            INC (j);
        END (*WHILE*);
        IF j <= HIGH(token) THEN
            token[j] := Nul;
        END (*IF*);
        IF j > 0 THEN
            Strings.Delete (line, 0, j);
        END (*IF*)
    END ParseToken;

(********************************************************************)

PROCEDURE ParseValue (VAR (*INOUT*) line: ARRAY OF CHAR;
                      VAR (*OUT*) value: ARRAY OF CHAR);

    (* A value is either a token or a quoted string. *)

    VAR pos: CARDINAL;  found: BOOLEAN;

    BEGIN
        IF line[0] = '"' THEN
            Strings.Delete (line, 0, 1);
            Strings.Assign (line, value);
            IF line[0] <> Nul THEN
                Strings.FindNext ('"', line, 0, found, pos);
                IF found THEN
                    IF pos <= HIGH(value) THEN
                        value[pos] := Nul;
                    END (*IF*);
                    Strings.Delete (line, 0, pos+1);
                ELSE
                    line[0] := Nul;
                END (*IF*);
            END (*IF*);
        ELSE
            ParseToken (line, value);
        END (*IF*);
    END ParseValue;

(********************************************************************)

PROCEDURE DeleteLeading (N: CARDINAL;  TS: TextStream;
                         VAR (*INOUT*) line: ARRAY OF CHAR;
                         VAR (*INOUT*) linecount, bytecount: CARDINAL;
                         VAR (*OUT*) overshot: BOOLEAN);

    (* Deletes the leading N characters of line, plus any space or  *)
    (* tab characters that follow.  If this takes us off the end of *)
    (* the line, we read in a new line and update linecount and     *)
    (* bytecount.                                                   *)

    (* Exception: if the newly read in line turns out not to be a   *)
    (* continuation line, then we don't update linecount and        *)
    (* bytecount, but instead set overshot = TRUE and leave the     *)
    (* line for the caller to look after.                           *)

    VAR length: CARDINAL;

    BEGIN
        overshot := FALSE;
        length := Strings.Length (line);
        LOOP
            WHILE (N < length)
                     AND ((line[N] = Space) OR (line[N] = Tab)) DO
                INC (N);
            END (*WHILE*);
            IF N = 0 THEN
                EXIT (*LOOP*);
            ELSE
                Strings.Delete (line, 0, N);
                DEC (length, N);
                N := 0;
                IF length = 0 THEN
                    TSReadLine (TS, line);
                    IF line[0] <> CtrlZ THEN
                        IF (line[0] = Space) OR (line[0] = Tab) THEN
                            INC (linecount);
                            INC (bytecount, LENGTH(line) + 2);
                        ELSE
                            overshot := TRUE;
                            EXIT (*LOOP*);
                        END (*IF*);
                    END (*IF*);
                    length := Strings.Length (line);
                END (*IF*);
            END (*IF*);
        END (*LOOP*);
    END DeleteLeading;

(********************************************************************)

PROCEDURE DeconstructContentType (TS: TextStream;
                        VAR (*INOUT*) line: ARRAY OF CHAR;
                        VAR (*OUT*) type, subtype: String32;
                        VAR (*OUT*) params: ParamPtr;
                        VAR (*INOUT*) subboundary: ARRAY OF CHAR;
                        VAR (*OUT*) linecount, bytecount: CARDINAL;
                        VAR (*OUT*) overshot: BOOLEAN);

    (* At this stage we have found a 'Content-type' header line,    *)
    (* and have deleted the keyword and colon and leading spaces.   *)
    (* Now we analyse the rest of that line, including continuation *)
    (* lines if any.  The output variables linecount and bytecount  *)
    (* give the amount of extra input we've read in.  They are      *)
    (* returned as zero if we didn't have to get any continuation   *)
    (* lines.  This procedure might in some cases overshoot to a    *)
    (* line that does not belong to us.  If so, we return with      *)
    (* overshot = TRUE and the unprocessed line available to the    *)
    (* caller.                                                      *)

    VAR previous, p: ParamPtr;

    BEGIN
        overshot := FALSE;
        linecount := 0;  bytecount := 0;
        params := NIL;  previous := NIL;
        subboundary[0] := Nul;
        ParseToken (line, type);
        IF line[0] = '/' THEN
            Strings.Delete (line, 0, 1);
            ParseToken (line, subtype);
        ELSE
            subtype := "";
        END (*IF*);
        DeleteLeading (0, TS, line, linecount, bytecount, overshot);
        WHILE (NOT overshot) AND (line[0] = ';') DO
            DeleteLeading (1, TS, line, linecount, bytecount, overshot);
            IF overshot THEN
                RETURN;
            END (*IF*);
            NEW (p);
            p^.next := NIL;
            ParseToken (line, p^.attribute);
            IF line[0] = '=' THEN
                DeleteLeading (1, TS, line, linecount, bytecount, overshot);
                IF overshot THEN
                    p^.value[0] := Nul;
                ELSE
                    ParseValue (line, p^.value);
                END (*IF*);
            ELSE
                p^.value[0] := Nul;
            END (*IF*);
            IF StringMatch(p^.attribute, "boundary") THEN
                Strings.Assign (p^.value, subboundary);
                Strings.Insert ("--", 0, subboundary);
            END (*IF*);
            IF params = NIL THEN
                params := p;
            ELSE
                previous^.next := p;
            END (*IF*);
            previous := p;
        END (*WHILE*);
    END DeconstructContentType;

(************************************************************************)
(*                      PARSING THE MIME STRUCTURE                      *)
(************************************************************************)

PROCEDURE BuildSD (TS: TextStream;
                   VAR (*INOUT*) stack: BoundaryStack;
                   VAR (*OUT*) boundarysize: CARDINAL;
                   TopLevel, IsAMessage, InDigest: BOOLEAN): StructureData;

    (* This is the internal version of BuildStructureData.              *)
    (* Parses the headers of a message file and returns the structure.  *)
    (* On exit, we have consumed the boundary code if any, and          *)
    (* boundarysize is the number of boundary code characters we have   *)
    (* actually consumed.  (For various reasons, the actual boundary    *)
    (* code can differ from that given on top of the boundary stack.)   *)

    (* Note: the boundary code actually belongs to the enclosing        *)
    (* section.  That means that, on exit, we have actually overshot    *)
    (* by 'boundarysize' characters in the text stream.  That doesn't   *)
    (* matter, because the caller expects us to overshoot.  The         *)
    (* IsAMessage case is special in that there is no boundary code.    *)

    (* TopLevel is TRUE iff this is the overall message, rather than    *)
    (* an embedded message part.  IsAMessage is TRUE iff this is either *)
    (* the top-level message or an embedded message/rfc822 part.        *)

    VAR linecount, bytecount, totallinecount, totalsize: CARDINAL;
        result: StructureData;
        line: ARRAY [0..2047] OF CHAR;
        subboundary: BoundaryString;
        HaveLine: BOOLEAN;

        (* SPECIAL NOTE: The HaveLine flag.                             *)

        (* Because we don't always know whether a header line is        *)
        (* followed by a continuation line, we will sometimes be in the *)
        (* situation where we have read, but not consumed, a lookahead  *)
        (* line.  HaveLine = TRUE iff the 'line' array contains a line  *)
        (* that we haven't yet finished processing.  It must be changed *)
        (* to FALSE by any part of the code that knows that the         *)
        (* processing of that line is complete.                         *)

    (********************************************************************)

    PROCEDURE GetNextLine;

        BEGIN
            IF NOT HaveLine THEN
                TSReadLine (TS, line);
                IF line[0] <> CtrlZ THEN
                    INC (linecount);
                    INC (totallinecount);
                    INC (bytecount, LENGTH(line) + 2);
                    INC (totalsize, LENGTH(line) + 2);
                END (*IF*);
                HaveLine := TRUE;
            END (*IF*);
        END GetNextLine;

    (********************************************************************)

    PROCEDURE DeleteLeading (N: CARDINAL);

        (* Deletes the leading N characters of line, plus any space or  *)
        (* tab characters that follow.  If this takes us off the end of *)
        (* the line, we read in a new line.                             *)

        VAR length: CARDINAL;

        BEGIN
            length := Strings.Length (line);
            LOOP
                WHILE (N < length)
                         AND ((line[N] = Space) OR (line[N] = Tab)) DO
                    INC (N);
                END (*WHILE*);
                IF N = 0 THEN
                    EXIT (*LOOP*);
                ELSE
                    Strings.Delete (line, 0, N);
                    DEC (length, N);
                    N := 0;
                    IF length = 0 THEN
                        HaveLine := FALSE;
                        GetNextLine;
                        length := Strings.Length (line);
                    END (*IF*);
                END (*IF*);
            END (*LOOP*);
        END DeleteLeading;

    (********************************************************************)

    PROCEDURE ParseContentType;

        (* At this stage we have found a 'Content-type' header line,    *)
        (* and have deleted the keyword and colon and leading spaces.   *)
        (* Now we analyse the rest of that line, including continuation *)
        (* lines if any.  Does not overshoot to a new input line.       *)

        VAR extralines, extrabytes: CARDINAL;

        BEGIN
            DeconstructContentType (TS, line, result^.Content_type,
                                    result^.Content_subtype,
                                    result^.firstparam, subboundary,
                                    extralines, extrabytes, HaveLine);
            INC (linecount, extralines);
            INC (totallinecount, extralines);
            INC (bytecount, extrabytes);
            INC (totalsize, extrabytes);
        END ParseContentType;

    (********************************************************************)

    PROCEDURE ProcessHeaderSection;

        (* Processes either a MIME header section or an RFC-2822 header *)
        (* section, pulling out the information we need.                *)
        (* We set HaveLine in some less-common cases.                   *)

        VAR InHeader: BOOLEAN;

        BEGIN
            InHeader := TRUE;
            HaveLine := FALSE;
            WHILE InHeader DO
                GetNextLine;
                HaveLine := FALSE;
                IF (line[0] = CtrlZ) OR (line[0] = Nul) THEN

                    (* The blank line, if found, counts as part of the header. *)

                    InHeader := FALSE;

                ELSE
                    (* The header lines that interest us all start with *)
                    (* the same substring, so we can save some time by  *)
                    (* doing a preliminary check.                       *)

                    IF Match (line, "Content-") THEN
                        IF HeaderMatch (line, "type") THEN
                            DeleteLeading (5);
                            ParseContentType;     (* might set HaveLine *)
                        ELSIF HeaderMatch (line, "transfer-encoding") THEN
                            DeleteLeading (18);
                            ParseToken (line, result^.Content_transfer_encoding);
                        ELSIF HeaderMatch (line, "ID") THEN
                            DeleteLeading (3);
                            ParseToken (line, result^.Content_ID);
                        ELSIF HeaderMatch (line, "description") THEN
                            DeleteLeading (12);
                            ParseToken (line, result^.Content_description);
                        END (*IF*);
                    END (*IF*);
                END (*IF*);
            END (*WHILE*);

            (* Insert defaults for missing fields. *)

            IF result^.Content_type[0] = Nul THEN
                IF InDigest THEN
                    result^.Content_type := "message";
                    result^.Content_subtype := "rfc822";
                ELSE
                    result^.Content_type := "text";
                    result^.Content_subtype := "plain";
                END (*IF*);
            END (*IF*);
            IF result^.Content_transfer_encoding[0] = Nul THEN
                result^.Content_transfer_encoding := "7bit";
            END (*IF*);

            (* This next bit compensates for a missing charset spec.    *)

            IF result^.firstparam = NIL THEN
                NEW (result^.firstparam);
                WITH result^.firstparam^ DO
                    next := NIL;
                    attribute := "charset";
                    value := "us-ascii";
                END (*WITH*);
            END (*IF*);

        END ProcessHeaderSection;

    (********************************************************************)

    VAR endofsection, partnum: CARDINAL;
        savedpos: FilePos;
        digest: BOOLEAN;
        tail, part: StructureData;
        top: BoundaryStack;
        <* IF testing THEN *>
            label: ARRAY [0..63] OF CHAR;
        <* END *>

    BEGIN
        (* Initialise all fields of the result. *)

        NEW (result);
        WITH result^ DO
            next := NIL;  parts := NIL;
            headerstart := TSCurrentPosition (TS);
            bodystart := headerstart;
            firstparam := NIL;
            multipart := FALSE;  ismessage := IsAMessage;
            is822part := FALSE;  NoMore := FALSE;
            envelope := NIL;
            bodysize := 0;  BodyLineCount := 0;
            totalsize := 0;  TotalLineCount := 0;
            Content_type[0] := Nul;
            Content_subtype[0] := Nul;
            Content_ID[0] := Nul;
            Content_description[0] := Nul;
            Content_transfer_encoding[0] := Nul;
        END (*WITH*);
        subboundary[0] := Nul;

        (****************************************************************)
        (*            ENVELOPE (FOR EMBEDDED MESSAGE ONLY)              *)
        (****************************************************************)

        (* For a message/rfc822 component, but apparently not for the   *)
        (* top level (the standard is ambiguous on this point), we have *)
        (* to construct the envelope.  If IsAMessage is TRUE then we    *)
        (* are now positioned at the start of the true message header,  *)
        (* not the MIME header, so we are about to read the envelope    *)
        (* details.  To avoid making the code too complex and obscure,  *)
        (* I am choosing to parse only the envelope fields here, and    *)
        (* then back up in the file to allow a second pass over the     *)
        (* header.  This second pass will pick up the "Content-xxx"     *)
        (* header information.                                          *)

        IF IsAMessage AND NOT TopLevel THEN
            result^.envelope := ParseEnvelope(TS);
            <* IF TESTING THEN *>
                Strings.Assign ("result^.envelope =", label);
                DebugDumpEnvelope (label, result^.envelope);
            <* END *>
            TSSetPosition (TS, result^.headerstart);
        END (*IF*);

        (****************************************************************)
        (*                 MAIL HEADER OR MIME HEADER                   *)
        (****************************************************************)

        linecount := 0;  bytecount := 0;
        totallinecount := 0; totalsize := 0;

        (* Process the message header section, if we're at the top      *)
        (* level, or the MIME header section, if we're in a nested part.*)

        ProcessHeaderSection;

        (* We have now passed the header lines, and are looking at      *)
        (* the body.  Reset the counts.                                 *)

        result^.headersize := bytecount;
        result^.bodystart := TSCurrentPosition (TS);
        linecount := 0;  bytecount := 0;

        (****************************************************************)
        (*                 BODY CASE 1: MULTIPART BODY                  *)
        (****************************************************************)

        result^.multipart := StringMatch (result^.Content_type, "MULTIPART");
        IF result^.multipart THEN

            digest := StringMatch (result^.Content_subtype, "DIGEST");

            (* The boundary code we are now looking for is the one in   *)
            (* subboundary.  (It was set while processing the header.)  *)
            (* Compensate for missing subboundary code.                 *)

            IF subboundary[0] = Nul THEN
                Strings.Assign (stack^.this, subboundary);
            END (*IF*);
            NEW (top);
            top^.next := stack;
            top^.this := subboundary;
            stack := top;

            (* Lines between the header and the first boundary code     *)
            (* should be ignored.  Note that GetNextLine is still       *)
            (* accumulating the line and byte counts.  We count this    *)
            (* ignored part as part of the body of the current message, *)
            (* but of course it won't be counted in the sizes of the    *)
            (* parts.                                                   *)

            REPEAT
                GetNextLine;
                HaveLine := FALSE;
            UNTIL AtBoundary(line, stack, boundarysize) > 0;

            tail := NIL;
            partnum := 1;
            REPEAT
                part := BuildSD (TS, stack, boundarysize, FALSE, FALSE, digest);

                INC (linecount, part^.TotalLineCount);
                INC (totallinecount, part^.TotalLineCount);
                INC (bytecount, part^.totalsize);
                INC (totalsize, part^.totalsize);

                (* The value of boundarysize at this point is the size  *)
                (* of the boundary between parts; this boundary is      *)
                (* part of the multipart structure, so should be        *)
                (* counted in our line and byte counts.                 *)

                IF boundarysize <> 0 THEN
                    INC (linecount);
                    INC (totallinecount);
                    INC (bytecount, boundarysize);
                    INC (totalsize, boundarysize);
                END (*IF*);

                IF tail = NIL THEN
                    result^.parts := part;
                ELSE
                    tail^.next := part;
                END (*IF*);
                tail := part;
                INC (partnum);

            UNTIL part^.NoMore;

            (* Lines after the last boundary code, the epilogue of this *)
            (* multipart part, should also be ignored.  By now the      *)
            (* boundary stack has been popped (we know this because     *)
            (* part^.NoMore is set).                                    *)
            (* The boundary we are now looking for belongs to the       *)
            (* caller, so we'll have to subtract its size from our own  *)
            (* calculated size.  This is done further down below.       *)

            REPEAT
                GetNextLine;
                HaveLine := FALSE;
                endofsection := AtBoundary(line, stack, boundarysize);
            UNTIL endofsection > 0;
            result^.NoMore := endofsection = 2;

        (****************************************************************)
        (*                 BODY CASE 2: MESSAGE/RFC822                  *)
        (****************************************************************)

        ELSIF StringMatch (result^.Content_type, "MESSAGE")
                  AND StringMatch (result^.Content_subtype, "RFC822") THEN

            (* A message/rfc822 component has both a MIME header and    *)
            (* a message header.  I am choosing here to represent it    *)
            (* by two tree levels.  Our result^ record records the size *)
            (* and location of the MIME header, and result^.parts       *)
            (* is for the actual embedded message.                      *)

            (* Note that the message will not be followed by a nested   *)
            (* boundary code; it is terminated by the parent's boundary *)
            (* code.  This is why we copy the "NoMore" code up (see     *)
            (* below).  To simplify the logic, we introduce a dummy     *)
            (* boundary code that should not match any line.            *)

            NEW (top);
            top^.next := stack;
            top^.this := "!dummy@#@separator!";
            stack := top;

            (* The dummy code cannot match any real boundary code,      *)
            (* because it does not start with "--".  It's also the sort *)
            (* of string that is unlikely to appear in body text.       *)

            part := BuildSD (TS, stack, boundarysize, FALSE, TRUE, FALSE);
            result^.is822part := TRUE;
            result^.parts := part;

            linecount := part^.TotalLineCount;
            INC (totallinecount, linecount);
            bytecount := part^.totalsize;
            INC (totalsize, bytecount);

            IF boundarysize <> 0 THEN
                INC (linecount);
                INC (totallinecount);
                INC (bytecount, boundarysize);
                INC (totalsize, boundarysize);
            END (*IF*);
            result^.NoMore := part^.NoMore;

        (****************************************************************)
        (*                BODY CASE 3: UNSTRUCTURED BODY                *)
        (****************************************************************)

        ELSE
            (* For the non-multipart case, all we have to do is count   *)
            (* the lines and characters.  Note that we have reset the   *)
            (* counts before entering this part of the code.            *)

            REPEAT
                GetNextLine;
                IF IsAMessage THEN
                    savedpos := TSCurrentPosition (TS);
                END (*IF*);
                endofsection := AtBoundary (line, stack, boundarysize);
                HaveLine := FALSE;
            UNTIL endofsection > 0;
            result^.NoMore := endofsection = 2;

        END (*IF*);

        (****************************************************************)
        (*             FINAL CORRECTION FOR ALL THREE CASES             *)
        (****************************************************************)

        (* Correct our size measurements for the final boundary *)
        (* overshoot.  An end of file gives boundarysize = 0.   *)
        (* An empty body can produce bytecount < boundarysize.  *)
        (* The backedup value also affects this calculation.    *)
        (* If backedup <>0 then this is also an overshoot.      *)

        IF boundarysize <> 0 THEN
            DEC (linecount);
            DEC (totallinecount);
            IF bytecount >= boundarysize THEN
                DEC (bytecount, boundarysize);
            ELSE
                bytecount := 0;
            END (*IF*);
            DEC (totalsize, boundarysize);
        END (*IF*);

        result^.BodyLineCount := linecount;
        result^.TotalLineCount := totallinecount;
        result^.bodysize := bytecount;
        result^.totalsize := totalsize;

        RETURN result;

    END BuildSD;

(************************************************************************)

PROCEDURE BuildStructureData (TS: TextStream;
                              VAR (*INOUT*) result: StructureData);

    (* Parses the headers of a message file and returns the structure. *)

    (* The result parameter is an INOUT parameter because I am         *)
    (* allowing for the possibility that the result has already been   *)
    (* worked out on a previous call.                                  *)

    VAR dummy: CARDINAL;
        stack, next: BoundaryStack;

    BEGIN
        IF result = NIL THEN
            NEW (stack);
            stack^.next := NIL;
            stack^.this[0] := CtrlZ;
            stack^.this[1] := Nul;
            result := BuildSD (TS, stack, dummy, TRUE, TRUE, FALSE);
            <* IF TESTING THEN *>
               DumpStructureData (result);
            <* END *>
            WHILE stack <> NIL DO
                next := stack^.next;
                DISPOSE (stack);
                stack := next;
            END (*WHILE*);
        END (*IF*);
    END BuildStructureData;

(************************************************************************)
(*                   PRODUCING A BODYSTRUCTURE STRING                   *)
(************************************************************************)

(*
        The following is an extract from the IMAP standard of the
        fields required in a BODYSTRUCTURE response.

             The basic fields of a non-multipart body part are
             in the following order:
             body type
                A string giving the content media type name as
                defined in [MIME-IMB].
             body subtype
                A string giving the content subtype name as
                defined in [MIME-IMB].
             body parameter parenthesized list
                A parenthesized list of attribute/value pairs
                [e.g. ("foo" "bar" "baz" "rag") where "bar" is
                the value of "foo" and "rag" is the value of
                "baz"] as defined in [MIME-IMB].
             body id
                A string giving the content id as defined in
                [MIME-IMB].
             body description
                A string giving the content description as
                defined in [MIME-IMB].
             body encoding
                A string giving the content transfer encoding as
                defined in [MIME-IMB].
             body size
                A number giving the size of the body in octets.
                Note that this size is the size in its transfer
                encoding and not the resulting size after any
                decoding.
             A body type of type MESSAGE and subtype RFC822
             contains, immediately after the basic fields, the
             envelope structure, body structure, and size in
             text lines of the encapsulated message.
             A body type of type TEXT contains, immediately
             after the basic fields, the size of the body in
             text lines.  Note that this size is the size in its
             content transfer encoding and not the resulting
             size after any decoding.
*)

(************************************************************************)

PROCEDURE AppendBodyStructure (SD: StructureData;  RC: ReplyCxt;
                               VAR (*INOUT*) buffer: ARRAY OF CHAR);

    (* Appends the StructureData information to buffer.  The parameter  *)
    (* RC permits us to send partial data if buffer is about to overflow. *)

    (********************************************************************)

    PROCEDURE PutField (VAR (*IN*) value: ARRAY OF CHAR);

        BEGIN
            IF value[0] = Nul THEN
                Strings.Append ('NIL', buffer);
            ELSE
                Strings.Append ('"', buffer);
                Strings.Append (value, buffer);
                Strings.Append ('"', buffer);
            END (*IF*);
        END PutField;

    (********************************************************************)

    VAR tempbuf: ARRAY [0..31] OF CHAR;
        p: StructureData;
        q: ParamPtr;

    BEGIN
        IF SD = NIL THEN
            Strings.Append ('NIL', buffer);
        ELSE
            Strings.Append ('(', buffer);
            IF SD^.multipart THEN

                p := SD^.parts;
                WHILE p <> NIL DO
                    AppendBodyStructure (p, RC, buffer);
                    p := p^.next;
                END (*WHILE*);
                Strings.Append (' ', buffer);
                PutField (SD^.Content_subtype);

            ELSE

                (* body type *)

                PutField (SD^.Content_type);

                (* body subtype *)

                Strings.Append (' ', buffer);
                PutField (SD^.Content_subtype);

                (* body parameter parenthesized list *)

                Strings.Append (' (', buffer);
                q := SD^.firstparam;
                WHILE q <> NIL DO
                    PutField (q^.attribute);
                    Strings.Append (' ', buffer);
                    PutField (q^.value);
                    q := q^.next;
                    IF q <> NIL THEN
                        Strings.Append (' ', buffer);
                    END (*IF*);
                END (*WHILE*);
                Strings.Append (') ', buffer);

                (* body id *)

                PutField (SD^.Content_ID);

                (* body description *)

                Strings.Append (' ', buffer);
                PutField (SD^.Content_description);

                (* body encoding *)

                Strings.Append (' ', buffer);
                PutField (SD^.Content_transfer_encoding);

                (* body size *)

                Strings.Append (' ', buffer);
                CardinalToStringLJ (SD^.bodysize, tempbuf);
                Strings.Append (tempbuf, buffer);

                (* Special cases: types TEXT and MESSAGE/RFC822 have some       *)
                (* extra fields beyond the basic ones above.                    *)

                IF StringMatch (SD^.Content_type, "TEXT") THEN

                    Strings.Append (' ', buffer);
                    CardinalToStringLJ (SD^.BodyLineCount, tempbuf);
                    Strings.Append (tempbuf, buffer);

                ELSIF StringMatch (SD^.Content_type, "MESSAGE")
                            AND StringMatch (SD^.Content_subtype, "RFC822") THEN

                    (* A body type of type MESSAGE and subtype RFC822       *)
                    (* contains, immediately after the basic fields, the    *)
                    (* envelope structure, body structure, and size in      *)
                    (* text lines of the encapsulated message.              *)

                    Strings.Append (' ', buffer);
                    AppendEnvelope (SD^.parts^.envelope, buffer);
                    PartialReply (RC, buffer);
                    buffer[0] := Nul;
                    Strings.Append (' ', buffer);
                    AppendBodyStructure (SD^.parts, RC, buffer);
                    PartialReply (RC, buffer);
                    buffer[0] := Nul;
                    Strings.Append (' ', buffer);
                    CardinalToStringLJ (SD^.parts^.TotalLineCount, tempbuf);
                    Strings.Append (tempbuf, buffer);

                END (*IF*);
            END (*IF*);

            Strings.Append (')', buffer);
        END (*IF*);

    END AppendBodyStructure;

(************************************************************************)

<* IF TESTING THEN *>

PROCEDURE WriteStructureOutline (file: ARRAY OF CHAR;
                                    VAR (*INOUT*) SD: StructureData);

    (* For testing: write an outline of the structure data. *)

    VAR TS: TextStream;

    BEGIN
        IF SD = NIL THEN
            TS := OpenForReading (file);
            IF TSFileOpened (TS) THEN
                BuildStructureData (TS, SD);
                CloseTS (TS);
            ELSE
                WriteString ("Can't open file ");  WriteString (file);
                WriteLn;
            END (*IF*);
        END (*IF*);
        IF SD <> NIL THEN
            DumpStructureData (SD);
        END (*IF*);
    END WriteStructureOutline;

<* END *>

(************************************************************************)
(*              MOVING TO A SPECIFIED PLACE IN A FILE                   *)
(************************************************************************)

PROCEDURE SkipToComponent (SD: StructureData;
                                   section: ARRAY OF CHAR): StructureData;

    (* The section number is either an empty string or a sequence of    *)
    (* numbers separated by dots, in the form 'section.subsection.etc'. *)
    (* We advance through SD until we come to the desired section.      *)
    (* On exit, the result points to a substructure of the original SD, *)
    (* depending on the "section" parameter.  We return NIL for a       *)
    (* nonexistent section.                                             *)

    VAR sectionnum, k: CARDINAL;  result: StructureData;

    BEGIN
        result := SD;
        WHILE (result <> NIL) AND (section[0] <> Nul) DO

            (* Decode the next number. *)

            sectionnum := 0;  k := 0;
            WHILE section[k] IN Digits DO
                sectionnum := 10*sectionnum + (ORD(section[k]) - ORD('0'));
                INC (k);
            END (*WHILE*);
            IF k > 0 THEN
                Strings.Delete (section, 0, k);
            END (*IF*);

            (* Move to the desired section.  Note: if we're not now in  *)
            (* a multipart component then we must already be in the     *)
            (* right section, because there aren't any other sections.  *)

            IF result^.multipart THEN
                result := result^.parts;
                WHILE sectionnum > 1 DO
                    IF result <> NIL THEN
                        result := result^.next;
                    END (*IF*);
                    DEC (sectionnum);
                END (*WHILE*);
            ELSIF sectionnum > 1 THEN
                result := NIL;
            END (*IF*);

            (* If we are now at a '.' we have to go down a level, and   *)
            (* this is covered in the code above.  In the is822part     *)
            (* case one level down in the numbering means two levels    *)
            (* down in the tree, so the extra step is done below.       *)

            IF section[0] = '.' THEN
                Strings.Delete (section, 0, 1);
                IF (result <> NIL) AND result^.is822part THEN
                    result := result^.parts;
                END (*IF*);
            END (*IF*);

        END (*WHILE*);

        RETURN result;

    END SkipToComponent;

(************************************************************************)

PROCEDURE SelectPart (TS: TextStream;  SD: StructureData;
                         section: ARRAY OF CHAR;  part: CARDINAL;
                                        VAR (*OUT*) size: CARDINAL);

    (* The section number is either an empty string or a sequence of    *)
    (* numbers separated by dots, in the form 'section.subsection.etc'. *)
    (* The part within that section is specified by the code:           *)
    (*          0   the entire section                                  *)
    (*          1   the MIME header                                     *)
    (*          2   the message header                                  *)
    (*          3   the text body of this section                       *)
    (* If the caller has specified something impossible, e.g. message   *)
    (* header for something that isn't a message, we default to our     *)
    (* best guess of what was intended.                                 *)
    (* On return, the position in the text stream TS has been set to    *)
    (* the beginning of the specified part, and the size parameter is   *)
    (* set to the size of that part.                                    *)

    VAR pos: FilePos;

    BEGIN
        (* Defaults: if looking for HEADER.FIELDS then go to header.    *)
        (* If nonexistent section then just set size = 0.               *)

        size := 0;   pos := FilePos{0,0};
        IF part > 3 THEN part := 2 END(*IF*);

        SD := SkipToComponent (SD, section);

        IF SD <> NIL THEN

            (* Finding the right starting point and size is obvious enough  *)
            (* in most cases, but we need to be careful in the case of an   *)
            (* internal message/rfc822 part, because that is represented    *)
            (* in two tree levels of the structure data: first the MIME     *)
            (* header, and then the actual message one level down.          *)

            IF SD^.is822part AND (part > 1) THEN
                SD := SD^.parts;
            END (*IF*);

            IF SD <> NIL THEN
                IF part = 0 THEN
                    size := SD^.totalsize;
                ELSIF part = 3 THEN
                    size := SD^.bodysize;
                ELSE
                    size := SD^.headersize;
                END (*IF*);

                IF part = 3 THEN
                    pos := SD^.bodystart;
                ELSE
                    pos := SD^.headerstart;
                END (*IF*);
            END (*IF*);
        END (*IF*);

        TSSetPosition (TS, pos);

    END SelectPart;

(************************************************************************)

END MailStructure.

