(**************************************************************************)
(*                                                                        *)
(*  IMAP add-on for the Weasel mail server                                *)
(*  Copyright (C) 2017   Peter Moylan                                     *)
(*                                                                        *)
(*  This program is free software: you can redistribute it and/or modify  *)
(*  it under the terms of the GNU General Public License as published by  *)
(*  the Free Software Foundation, either version 3 of the License, or     *)
(*  (at your option) any later version.                                   *)
(*                                                                        *)
(*  This program is distributed in the hope that it will be useful,       *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU General Public License for more details.                          *)
(*                                                                        *)
(*  You should have received a copy of the GNU General Public License     *)
(*  along with this program.  If not, see <http://www.gnu.org/licenses/>. *)
(*                                                                        *)
(*  To contact author:   http://www.pmoylan.org   peter@pmoylan.org       *)
(*                                                                        *)
(**************************************************************************)

IMPLEMENTATION MODULE Messages;

        (********************************************************)
        (*                                                      *)
        (*         Properties of messages at the file level     *)
        (*                                                      *)
        (*  Programmer:         P. Moylan                       *)
        (*  Started:            16 March 2003                   *)
        (*  Last edited:        5 July 2017                     *)
        (*  Status:                                             *)
        (*           Undergoing a major overhaul.               *)
        (*                                                      *)
        (*     BODY[part]<partial> implemented but untested.    *)
        (*     We now separate the fetch into two phases:       *)
        (*     working out the MIME structure of the message,   *)
        (*     and then using that analysis to locate the       *)
        (*     position and size of the required section.       *)
        (*                                                      *)
        (*     Now resuming this job after a long time gap,     *)
        (*     so I'll have to re-check my design.              *)
        (*                                                      *)
        (*     Update 31-Aug-2009: BODY[]<partial> is working   *)
        (*     except for the fact that it seems to be          *)
        (*     returning too few bytes.  (10009 instead of      *)
        (*     10240, in the example I'm working on, so at      *)
        (*     least it's the right order of magnitude.)        *)
        (*                                                      *)
        (*     Procedure SendFragmentedFilePart needs more work *)
        (*     to cover the (unlikely but possible) "partial"   *)
        (*     case.                                            *)
        (*                                                      *)
        (********************************************************)

(************************************************************************)
(*                                                                      *)
(*  ISSUES THAT NEED TO BE RESOLVED                                     *)
(*                                                                      *)
(*  1.  The present version includes a lot of redundant recalculation   *)
(*      of body structure.  Ultimately we need some sort of cache for   *)
(*      such data, with a check on whether the data structure already   *)
(*      exists.  (Probably in module MailStructure, but an issue for    *)
(*      this module is whether we should delete the structure data      *)
(*      after using it.)  For now, we have to give a greater priority   *)
(*      to correctness rather than efficiency, so the cacheing issue    *)
(*      can be delayed until we have done more testing.                 *)
(*                                                                      *)
(*  2.  Our BodyRequestInfo data structure has an implicit assumption   *)
(*      that only certain combinations of FETCH parameters can occur.   *)
(*      For example, we assume that things like FETCH 1.MIME 3.HEADER   *)
(*      will not occur in a single operation.  In fact the standard     *)
(*      seems to allow combinations that would require multiple passes  *)
(*      over the same file, so we should probably be allowing for the   *)
(*      possibility of an arbitrarily complicated sequence in a single  *)
(*      FETCH operation.  Maybe it will be sufficient to just build a   *)
(*      linked list of BodyRequestInfo records.                         *)
(*                                                                      *)
(************************************************************************)

(*******************)
(* Debugging stuff *)
(*******************)

FROM IOChan IMPORT ChanId;

(*******************)

FROM SYSTEM IMPORT ADDRESS, CAST;

FROM TransLog IMPORT
    (* type *)  TransactionLogID,
    (* proc *)  LogTransaction;

IMPORT Strings, FileSys;

FROM BufferedTextStreams IMPORT
    (* type *)  TextStream,
    (* proc *)  OpenForReading, CloseTS, TSReadRaw, TSReadLine,
                TSStartPosition, TSCurrentPosition, TSSetPosition;

FROM FileOps IMPORT
    (* type *)  DirectoryEntry, FilePos,
    (* proc *)  GetFileSize, OpenOldFile, CloseFile, OpenNewFile1,
                FWriteString, FWriteLn, FirstDirEntry, DirSearchDone,
                SetFileSize, WriteRaw, DeleteFile, ReadRaw;

FROM Parser IMPORT
    (* proc *)  Date, PackDate, Match, SPACE, AString, Number;

FROM Replies IMPORT
    (* type *)  ReplyCxt,
    (* proc *)  Reply, PartialReply, CommitReply, LogMessage,
                SuppressLogging, RestoreLogging;

FROM Envelopes IMPORT
    (* type *)  EnvelopePtr,
    (* proc *)  ParseEnvelope, AppendEnvelope, DiscardEnvelope;

FROM MailStructure IMPORT
    (* type *)  StructureData,
    (* proc *)  BuildStructureData, DiscardStructureData, AppendBodyStructure,
                SelectPart;

FROM Conversions IMPORT
    (* proc *)  CardinalToStringLJ;

FROM Names IMPORT
    (* type *)  FilenameString, ArgPointer;

FROM MiscFuncs IMPORT
    (* proc *)  StringMatch;

FROM LowLevel IMPORT
    (* proc *)  EVAL;

FROM Types IMPORT
    (* type *)  CARD64;

FROM LONGLONG IMPORT
    (* proc *)  Add64, Sub64, Diff64, ORDL;

FROM Storage IMPORT
    (* proc *)  ALLOCATE, DEALLOCATE;

(************************************************************************)

CONST
    Nul = CHR(0);  Tab = CHR(9);  Space = ' ';  CtrlZ = CHR(26);

TYPE
    CharSet = SET OF CHAR;

CONST
    Digits = CharSet {'0'..'9'};

TYPE
    Param = ARRAY [0..511] OF CHAR;
    ParamList = POINTER TO
                    RECORD
                        next: ParamList;
                        text: Param;
                    END (*RECORD*);

    (* BodyRequestInfo records what is being requested in a BODY[]      *)
    (* FETCH request.  The standard leaves it ambiguous as to whether   *)
    (* multiple parts can be requested in a single fetch.  (The syntax  *)
    (* permits it, but it's hard to tell whether that is what was       *)
    (* intended.)  For now I'm allowing for a single part, which means  *)
    (* that the "BodyParts" type definition is likely to be scrapped.   *)
    (* In the longer term it might be necessary to move to the          *)
    (* MultiBodyRequestInfo linked list structure which is declared     *)
    (* below.  If so, the parsing in module Boxes will probably also    *)
    (* have to be modified.  All that is certain for now is that some   *)
    (* of the following declarations are redundant; but I don't yet     *)
    (* know which ones.                                                 *)

    (*BodyPart = (HEADER_FIELDS, HEADER_FIELDS_NOT, HEADER, MIME, TEXT, ALL);*)
    BodyPart = (ALL, MIME, HEADER, TEXT, HEADER_FIELDS, HEADER_FIELDS_NOT);
    BodyParts = SET OF BodyPart;
    BodyRequestInfo = POINTER TO RECORD
                   section: ARRAY [0..63] OF CHAR;
                   part: BodyPart;
                   parts: BodyParts;
                   Headers: ParamList;
                   partial: BOOLEAN;
                   initialoffset, amount: CARDINAL;

                   (* Note: if partial is FALSE then initialoffset = 0  *)
                   (* and amount = MAX(CARDINAL).                       *)

               END (*RECORD*);

    (* New version of the above - which I will probably scrap. *)

    OneBodyRequestInfo = POINTER TO RECORD
                      section: ARRAY [0..63] OF CHAR;
                      part: BodyPart;
                      Headers: ParamList;
                      partial: BOOLEAN;
                      initialoffset, amount: CARDINAL;
                  END (*RECORD*);

    MultiBodyRequestInfo = POINTER TO MultiRecord;
    MultiRecord = RECORD
                      this: OneBodyRequestInfo;
                      next: MultiBodyRequestInfo;
                  END (*RECORD*);

(************************************************************************)
(*                          MESSAGE SIZE                                *)
(************************************************************************)

PROCEDURE SizeOfMessage (VAR (*IN*) file: FilenameString): CARD64;

    (* Returns the RFC822.SIZE attribute. *)

    BEGIN
        RETURN GetFileSize (file);
    END SizeOfMessage;

(************************************************************************)
(*                            DATE OPERATIONS                           *)
(************************************************************************)

PROCEDURE PackedDate (VAR (*IN*) file: FilenameString): CARDINAL;

    (* Returns the date (in packed format) that this file was received. *)

    VAR D: DirectoryEntry;
        packeddate: CARDINAL;

    BEGIN
        IF FirstDirEntry (file, FALSE, TRUE, D) THEN
            packeddate := D.datePkd;
        ELSE
            packeddate := 0;
        END (*IF*);
        DirSearchDone (D);
        RETURN packeddate;
    END PackedDate;

(************************************************************************)

PROCEDURE PackedDateTime (VAR (*IN*) file: FilenameString;  VAR (*OUT*) date, time: CARDINAL);

    (* Returns the date and time (in packed format) that this file was received. *)

    VAR D: DirectoryEntry;

    BEGIN
        IF FirstDirEntry (file, FALSE, TRUE, D) THEN
            date := D.datePkd;
            time := D.timePkd;
        ELSE
            date := 0;
            time := 0;
        END (*IF*);
        DirSearchDone (D);
    END PackedDateTime;

(************************************************************************)

PROCEDURE HeaderMatch (VAR (*IN*) line: ARRAY OF CHAR;
                                header: ARRAY OF CHAR): BOOLEAN;  FORWARD;

(************************************************************************)

PROCEDURE GetHeaderText (file: FilenameString;  key: ARRAY OF CHAR;
                            VAR (*OUT*) text: ARRAY OF CHAR): BOOLEAN;

    (* Sets text to the string following 'key:' in the header lines.  *)

    VAR InHeader, found: BOOLEAN;
        TS: TextStream;
        line: ARRAY [0..2047] OF CHAR;

    BEGIN
        found := FALSE;  InHeader := TRUE;
        TS := OpenForReading (file);
        WHILE InHeader AND NOT found DO
            TSReadLine (TS, line);
            IF (line[0] = CtrlZ) OR (line[0] = Nul) THEN
                InHeader := FALSE;
            ELSIF HeaderMatch (line, key) THEN
                Strings.Assign (line, text);
                Strings.Delete (text, 0, LENGTH(key)+1);
                found := TRUE;
            END (*IF*);
        END (*WHILE*);
        CloseTS (TS);
        RETURN found;
    END GetHeaderText;

(************************************************************************)

PROCEDURE HeaderDate (VAR (*IN*) file: FilenameString): CARDINAL;

    (* Returns the date (in packed format) from the 'Date:' header      *)
    (* line.  If there is no such line, date is 00 Jan 1980.            *)

    VAR result, pos: CARDINAL;
        text: ARRAY [0..2047] OF CHAR;
        datestring: ArgPointer;

    BEGIN
        IF GetHeaderText (file, 'Date', text) THEN
            pos := 0;
            WHILE (pos < 2048) AND NOT (text[pos] IN Digits) DO
                INC (pos);
            END (*WHILE*);
            IF pos > 0 THEN
                Strings.Delete (text, 0, pos);
            END (*IF*);
            NEW (datestring);
            IF Date (text, datestring^) THEN
                result := PackDate (datestring);
            ELSE
                result := 0;
            END (*IF*);
            DISPOSE (datestring);
        ELSE
            result := 0;
        END (*IF*);
        RETURN result;
    END HeaderDate;

(************************************************************************)
(*                             SEARCHES                                 *)
(*  Note that these do not require any parsing of the MIME structure,   *)
(*  since for our present searches it is sufficient to treat a message  *)
(*  as something of the form (header)+(everything else).                *)
(************************************************************************)

PROCEDURE HeaderMatch (VAR (*IN*) line: ARRAY OF CHAR;
                                  header: ARRAY OF CHAR): BOOLEAN;

    (* Returns TRUE iff header':' is a leading substring of line. *)

    VAR N: CARDINAL;
        test: ARRAY [0..1023] OF CHAR;

    BEGIN
        N := Strings.Length (header);

        (* Check for the colon first, because this will eliminate  *)
        (* most cases.                                             *)

        IF line[N] = ':' THEN
            Strings.Assign (line, test);
            test[N] := Nul;
            Strings.Capitalize (test);
            Strings.Capitalize (header);
            RETURN Strings.Equal (test, header);
        ELSE
            RETURN FALSE;
        END (*IF*);

    END HeaderMatch;

(************************************************************************)

PROCEDURE HeaderListMatch (VAR (*IN*) line: ARRAY OF CHAR;
                               VAR (*INOUT*) Headers: ParamList): BOOLEAN;

    (* Returns TRUE iff header':' is a leading substring of line, where *)
    (* header is one element of Headers.  We don't delete any of the    *)
    (* list, because it could be used more than once.                   *)

    VAR this: ParamList;
        success: BOOLEAN;

    BEGIN
        success := FALSE;  this := Headers;
        WHILE NOT success AND (this <> NIL) DO
            IF HeaderMatch (line, this^.text) THEN
                success := TRUE;
            ELSE
                this := this^.next;
            END (*IF*);
        END (*WHILE*);
        RETURN success;
    END HeaderListMatch;

(************************************************************************)

PROCEDURE MatchText (skip: CARDINAL;
                           VAR (*IN*) line, text: ARRAY OF CHAR): BOOLEAN;

    (* Returns TRUE iff text is a substring of line.  Before checking   *)
    (* we skip the first 'skip' characters of line.                     *)

    VAR pos: CARDINAL;
        found: BOOLEAN;

    BEGIN
        Strings.Capitalize (line);
        Strings.Capitalize (text);
        Strings.FindNext (text, line, skip, found, pos);
        RETURN found;
    END MatchText;

(************************************************************************)

PROCEDURE FindInHeader (VAR (*IN*) file: FilenameString;
                            VAR (*IN*) key, text: ARRAY OF CHAR): BOOLEAN;

    (* Returns TRUE iff there is a header line labelled 'key:'  *)
    (* that has text as a substring.                            *)

    VAR InHeader, found, keyfound: BOOLEAN;
        TS: TextStream;
        line: ARRAY [0..2047] OF CHAR;

    BEGIN
        found := FALSE;  InHeader := TRUE;  keyfound := FALSE;
        TS := OpenForReading (file);
        WHILE InHeader AND NOT found DO
            TSReadLine (TS, line);
            IF (line[0] = CtrlZ) OR (line[0] = Nul) THEN
                InHeader := FALSE;
            ELSIF (line[0] = Space) OR (line[0] = Tab) THEN
                IF keyfound THEN
                    found := MatchText (1, line, text);
                END (*IF*);
            ELSIF HeaderMatch (line, key) THEN
                keyfound := TRUE;
                found := MatchText (Strings.Length(key)+1, line, text);
            ELSE
                keyfound := FALSE;
            END (*IF*);
        END (*WHILE*);
        CloseTS (TS);
        RETURN found;
    END FindInHeader;

(************************************************************************)

PROCEDURE FindInText (VAR (*IN*) file: FilenameString;
                        VAR (*IN*) string: ARRAY OF CHAR;
                                 IncludeHeader: BOOLEAN): BOOLEAN;

    (* Returns TRUE iff the file includes the given string      *)
    (* (case insensitive).  Searches the header lines as well   *)
    (* as the body iff IncludeHeader is TRUE.                   *)

    VAR found, InBody, EOF: BOOLEAN;
        TS: TextStream;
        line: ARRAY [0..2047] OF CHAR;

    BEGIN
        found := FALSE;  InBody := FALSE;  EOF := FALSE;
        TS := OpenForReading (file);
        REPEAT
            TSReadLine (TS, line);
            IF line[0] = CtrlZ THEN
                EOF := TRUE;
            ELSIF line[0] = Nul THEN
                InBody := TRUE;
            ELSIF InBody OR IncludeHeader THEN
                found := MatchText (1, line, string);
            END (*IF*);
        UNTIL found OR EOF;
        CloseTS (TS);
        RETURN found;
    END FindInText;

(************************************************************************)
(*                               ENVELOPE                               *)
(************************************************************************)

PROCEDURE ReturnEnvelope (VAR (*IN*) file: FilenameString;
                               VAR (*INOUT*) buffer: ARRAY OF CHAR);

    (* Appends the ENVELOPE information to buffer.  Note that this is   *)
    (* for the top-level information for the entire message, and can    *)
    (* be worked out without working out the MIME structure of the      *)
    (* message.  An ENVELOPE request never asks for the envelope of an  *)
    (* embedded rfc-822 part.  That embedded data is, however, needed   *)
    (* for a FETCH that asks for BODY or BODYSTRUCTURE information.     *)

    VAR EnvData: EnvelopePtr;
        TS: TextStream;

    BEGIN
        TS := OpenForReading (file);
        EnvData := ParseEnvelope(TS);
        CloseTS (TS);
        AppendEnvelope (EnvData, buffer);
        DiscardEnvelope (EnvData);
    END ReturnEnvelope;

(************************************************************************)
(*                          BODY REQUEST INFO                           *)
(*                                                                      *)
(*   A BodyRequestInfo structure records what the client is requesting  *)
(*   in terms of body parts.  To satisfy the request we first parse the *)
(*   request to get a BodyRequestInfo object, then parse the message to *)
(*   get a StructureData object; and then we match up the two to work   *)
(*   out what to send back.                                             *)
(*                                                                      *)
(************************************************************************)

PROCEDURE InitBodyRequestInfo (VAR (*OUT*) BI: BodyRequestInfo);

    (* Sets BI to an initially clear state. *)

    BEGIN
        BI := NIL;
    END InitBodyRequestInfo;

(************************************************************************)

PROCEDURE DiscardParamList (VAR (*INOUT*) PL: ParamList);

    (* Disposes of a parameter list. *)

    VAR this: ParamList;

    BEGIN
        WHILE PL <> NIL DO
            this := PL;
            PL := PL^.next;
            DISPOSE (this);
        END (*WHILE*);
    END DiscardParamList;

(************************************************************************)

PROCEDURE DiscardBodyRequestInfo (VAR (*INOUT*) BI: BodyRequestInfo);

    (* Disposes of a BodyRequestInfo structure. *)

    BEGIN
        IF BI <> NIL THEN
            DiscardParamList (BI^.Headers);
            DISPOSE (BI);
        END (*WHILE*);
    END DiscardBodyRequestInfo;

(************************************************************************)

PROCEDURE ParseFetchSection (VAR (*IN*) params: ARRAY OF CHAR;
                             VAR (*OUT*) result: BodyRequestInfo);

    (* Parses the BODY[<section>] (with an optional <<partial>>) part   *)
    (* of the FETCH command parameters.  On entry the caller has        *)
    (* already consumed the leading '['.  Note: this is the command     *)
    (* parsing.  The actual fetching is handled elsewhere.  Here we're  *)
    (* simply figuring out what we're supposed to fetch.                *)

    (********************************************************************)

    PROCEDURE ParseHeaderList;

        (* Puts a list of header names into result^.Headers.  *)

        VAR prev, this: ParamList;
            thisheader: Param;

        BEGIN
            WHILE SPACE(params) DO
                (* skip space characters *)
            END (*WHILE*);
            prev := NIL;  result^.Headers := NIL;
            IF Match (params, "(") THEN
                WHILE (params[0] <> Nul) AND (params[0] <> ")")
                               AND AString (params, thisheader) DO
                    NEW (this);
                    this^.next := NIL;
                    this^.text := thisheader;
                    IF prev = NIL THEN
                        result^.Headers := this;
                    ELSE
                        prev^.next := this;
                    END (*IF*);
                    prev := this;
                    WHILE SPACE(params) DO
                        (* skip space characters *)
                    END (*WHILE*);
                END (*WHILE*);
                EVAL (Match (params, ")"));
            END (*IF*);
        END ParseHeaderList;

    (********************************************************************)

    TYPE CharSet = SET OF CHAR;
    CONST ExtDigits = CharSet {'0'..'9', '.'};

    VAR k: CARDINAL;  ch: CHAR;

    BEGIN
        NEW (result);
        WITH result^ DO

            (* A part specification has the form N.N.N.part, where N    *)
            (* is a number (there can be more or less than three of     *)
            (* them, and part is a string like "MIME" or "HEADER" (see  *)
            (* below for the possibilities).  Here we put the N.N.N bit *)
            (* in section, and a code for the final string in parts.    *)
            (* It is legal to have a null string for section.           *)

            (* We don't bother reporting excess periods, although that  *)
            (* is technically a client syntax error.  We simply discard *)
            (* them silently, so that software elswhere in this module  *)
            (* can treat a period is section as an unambiguous "down    *)
            (* one level" specification.                                *)

            k := 0;
            ch := params[0];
            WHILE ch IN ExtDigits DO
                section[k] := ch;
                INC (k);
                WHILE (ch = '.') AND (params[k] = '.') DO
                    (* Delete excess periods. *)
                    INC (k);
                END (*WHILE*);
                ch := params[k];
            END (*IF*);
            section[k] := Nul;
            IF k > 0 THEN
                Strings.Delete (params, 0, k);
                IF section[k-1] = '.' THEN
                    (* Delete trailing period. *)
                    section[k-1] := Nul;
                END (*IF*);
            END (*IF*);

            result^.Headers := NIL;
            IF Match (params, "HEADER.FIELDS.NOT") THEN
                part := HEADER_FIELDS_NOT;
                INCL (parts, part);
                ParseHeaderList;
            ELSIF Match (params, "HEADER.FIELDS") THEN
                part := HEADER_FIELDS;
                INCL (parts, part);
                ParseHeaderList;
            ELSIF Match (params, "HEADER") THEN
                part := HEADER;
                INCL (parts, part);
                EXCL (parts, HEADER_FIELDS);
                EXCL (parts, HEADER_FIELDS_NOT);
            ELSIF Match (params, "MIME") THEN
                part := MIME;
                INCL (parts, part);
            ELSIF Match (params, "TEXT") THEN
                part := TEXT;
                INCL (parts, part);
            ELSE

                (* Empty section specification implies the entire   *)
                (* message if this part is a message, or all except *)
                (* the MIME header if this part is a part of a      *)
                (* multipart message.                               *)

                part := ALL;
                INCL (parts, part);

            END (*IF*);

            (* Consume the closing bracket, and then check for the  *)
            (* <partial> option.                                    *)

            EVAL (Match (params, ']'));
            IF Match (params, '<') THEN
                partial := TRUE;
                EVAL (Number (params, initialoffset));
                EVAL (Match (params, '.'));
                EVAL (Number (params, amount));
                EVAL (Match (params, '>'));
            ELSE
                initialoffset := 0;
                amount := MAX(CARDINAL);
            END (*IF*);

        END (*WITH*);

    END ParseFetchSection;

(************************************************************************)
(*                      RETURNING THE BODY STRUCTURE                    *)
(************************************************************************)

PROCEDURE ReportBodyStructure (VAR (*IN*) file: FilenameString;
                               RC: ReplyCxt;
                               VAR (*INOUT*) buffer: ARRAY OF CHAR;
                               VAR (*INOUT*) SD: StructureData);

    (* Creates the BODYSTRUCTURE reply for a fetch, and appends *)
    (* it to buffer.  The RC parameter permits us to send a     *)
    (* partial reply when the buffer is about to overflow.      *)
    (* The SD parameter allows us to work out the structure     *)
    (* once and preserve the answer over several calls, but in  *)
    (* the present version we destroy the structure data after  *)
    (* using it as a guard against memory leaks.                *)

    VAR TS: TextStream;

    BEGIN
        TS := OpenForReading (file);
        BuildStructureData (TS, SD);
        CloseTS (TS);
        AppendBodyStructure (SD, RC, buffer);
        DiscardStructureData (SD);
    END ReportBodyStructure;

(************************************************************************)
(*                   EXTRACTING SECTIONS OF A MESSAGE                   *)
(************************************************************************)

PROCEDURE SendHeaderFields (info: BodyRequestInfo;
                               TS: TextStream;
                               offset, limit: CARDINAL;
                               RC: ReplyCxt; sendthem: BOOLEAN): CARDINAL;

    (* This function is called only when info^.part is HEADER_FIELDS    *)
    (* or HEADER_FIELDS_NOT.  If sendthem is FALSE, returns the number  *)
    (* of characters that should be sent.  IF sendthem is TRUE, sends   *)
    (* those characters.  We assume that the caller has already set     *)
    (* the text stream position to the start of the header.             *)

    CONST buffersize = 2048;

    VAR size, excess, L: CARDINAL;
        InHeader, PastOffset, IncludeIt: BOOLEAN;
        line: ARRAY [0..buffersize-1] OF CHAR;

    BEGIN
        size := 0;
        InHeader := TRUE;  IncludeIt := FALSE;
        PastOffset := offset = 0;
        WHILE InHeader DO

            TSReadLine (TS, line);
            InHeader := line[0] <> Nul;
            IF NOT InHeader THEN
                INC (size, 2);
                IncludeIt := FALSE;
                IF sendthem THEN
                    Reply (RC, "");
                END (*IF*);

            (* We deal with continuation lines by simply leaving    *)
            (* the IncludeIt flag at its previous value.            *)

            ELSIF (line[0] <> Space) AND (line[0] <> Tab) THEN

                IncludeIt := FALSE;
                IF (info^.part = HEADER) OR (info^.part = ALL) THEN
                    IncludeIt := TRUE;
                ELSIF info^.part = HEADER_FIELDS THEN
                    IncludeIt := HeaderListMatch (line, info^.Headers);
                ELSIF info^.part = HEADER_FIELDS_NOT THEN
                    IncludeIt := NOT HeaderListMatch (line, info^.Headers);
                END (*IF*);
            END (*IF*);

            IF IncludeIt THEN
                L := Strings.Length(line);
                INC (size, L + 2);
                IF PastOffset THEN
                    IF size > limit THEN
                        excess := size - limit;
                        IF excess <= L THEN
                            line[L-excess] := Nul;
                            excess := 0;
                        ELSE
                            line[0] := Nul;
                        END (*IF*);
                        size := limit;
                        InHeader := FALSE;
                    ELSE
                        excess := 0;
                    END (*IF*);
                    IF sendthem AND (excess = 0) THEN
                        Reply (RC, line);
                    END (*IF*);
                ELSE
                    (* We're still looking for the starting point. *)

                    IF size >= offset THEN

                        (* Restart the size calculation. *)

                        PastOffset := TRUE;
                        IF sendthem AND (size >= offset+2) THEN
                            Strings.Delete (line, 0, size-offset-2);
                            Reply (RC, line);
                        END (*IF*);
                        DEC (size, offset);
                    END (*IF*);

                END (*IF*);
            END (*IF*);

        END (*WHILE*);

        RETURN size;

    END SendHeaderFields;

(************************************************************************)

PROCEDURE SendFragmentedFilePart (RC: ReplyCxt;  TS0: TextStream;
                                   info: BodyRequestInfo;
                                      SD: StructureData);

    (* Sends to the output channel that contains the subset of TS0      *)
    (* specified by the 'info' parameter.  This procedure is invoked    *)
    (* only in the special case where info^.part is equal to either     *)
    (* HEADER_FIELDS or HEADER_FIELDS_NOT.  These two cases require     *)
    (* special treatment because what we are going to send is not       *)
    (* necessarily a contiguous part of the file, and two passes are    *)
    (* needed to do the job properly.                                   *)

    CONST buffersize = 2048;

    VAR bytestosend, size: CARDINAL;
        pos: FilePos;
        sizebuff: ARRAY [0..15] OF CHAR;
        boundary: ARRAY [0..buffersize-1] OF CHAR;

    BEGIN
        (* Locate the beginning of the desired section. *)

        boundary[0] := CtrlZ;  boundary[1] := Nul;
        SelectPart (TS0, SD, info^.section, 2, size);
        pos := TSCurrentPosition (TS0);

        (* At this stage SD^.headerstart says where the desired section *)
        (* starts, and SD^.BodyByteCount and SD^.TotalByteCount give    *)
        (* the size of the body and the complete part, as appropriate.  *)
        (* Do an initial pass over the header to work out the size of   *)
        (* the requested information.                                   *)

        IF size > 0 THEN
            bytestosend := SendHeaderFields (info, TS0,
                                       info^.initialoffset, info^.amount,
                                       RC, FALSE);
        ELSE
            bytestosend := 0;
        END (*IF*);

        (* Go back to the beginning of the desired section. *)

        TSSetPosition (TS0, pos);

        (* Start sending the required body section. *)

        PartialReply (RC, " {");
        CardinalToStringLJ (bytestosend, sizebuff);
        PartialReply (RC, sizebuff);
        PartialReply (RC, "}");
        CommitReply (RC);

        (* Send the specified header lines.      *)

        LogMessage (RC, "  ... content suppressed in log ... )");
        SuppressLogging (RC);
        IF bytestosend > 0 THEN
            EVAL (SendHeaderFields (info, TS0,
                                       info^.initialoffset, info^.amount,
                                       RC, TRUE));
        END (*IF*);

    END SendFragmentedFilePart;

(************************************************************************)

PROCEDURE SendContiguousFilePart (RC: ReplyCxt;  TS0: TextStream;
                                   info: BodyRequestInfo;
                                       SD: StructureData);

    (* Sends to the output channel that contains the subset of TS0      *)
    (* specified by the 'info' parameter.  This procedure is invoked    *)
    (* in all cases except where info^.part is equal to one of          *)
    (* HEADER_FIELDS or HEADER_FIELDS_NOT.                              *)

    CONST buffersize = 2048;

    VAR togo, amount, actual, bytestosend, code: CARDINAL;
        sizebuff: ARRAY [0..15] OF CHAR;
        start: FilePos;
        line, boundary: ARRAY [0..buffersize-1] OF CHAR;

    VAR bytessent: CARDINAL;

    BEGIN
        (* Locate the beginning of the desired section. *)

        boundary[0] := CtrlZ;  boundary[1] := Nul;
        CASE info^.part OF
            ALL:        code := 0;
          | MIME:       code := 1;
          | HEADER:     code := 2;
          | TEXT:       code := 3;
        END (*CASE*);
        SelectPart (TS0, SD, info^.section, code, bytestosend);
        start := TSCurrentPosition (TS0);

        IF info^.partial THEN

            (* May need to retain only part of the file.  Up until now, *)
            (* the value of bytestosend is the maximum number of bytes  *)
            (* available in the specified section.  If we move the      *)
            (* start point forward, we must adjust the available byte   *)
            (* count accordingly.                                       *)

            Add64 (start, info^.initialoffset);
            IF bytestosend > info^.initialoffset THEN
                DEC (bytestosend, info^.initialoffset);
            ELSE
                bytestosend := 0;
            END (*IF*);

            (* Set bytestosend equal to the maximum of the amount       *)
            (* available and the amount requested.                      *)

            IF info^.amount < bytestosend THEN
                bytestosend := info^.amount;
            END (*IF*);

        END (*IF*);

        (* Start sending the required body section. *)

        PartialReply (RC, " {");
        CardinalToStringLJ (bytestosend, sizebuff);
        PartialReply (RC, sizebuff);
        PartialReply (RC, "}");
        CommitReply (RC);

        TSSetPosition (TS0, start);
        LogMessage (RC, "  ... content suppressed in log ... )");
        SuppressLogging (RC);

        (* Copy the data from the current position. *)

        bytessent := 0;
        togo := bytestosend;
        WHILE togo > 0 DO
            amount := buffersize;
            IF togo < buffersize THEN
                amount := togo;
            END (*IF*);
            TSReadRaw (TS0, line, amount, actual);
            IF actual = 0 THEN
                togo := 0;
            ELSE
                IF actual < buffersize THEN
                    line[actual] := Nul;
                END (*IF*);
                PartialReply (RC, line);
                DEC (togo, actual);
                INC (bytessent, actual);
            END (*IF*);
        END (*WHILE*);

    END SendContiguousFilePart;

(************************************************************************)

PROCEDURE SendFilePart (RC: ReplyCxt;  file: FilenameString;
                               info: BodyRequestInfo;  SD: StructureData);

    (* Sends to the output channel that contains the subset of 'file'   *)
    (* specified by the 'info' parameter.                               *)

    VAR TS: TextStream;

    BEGIN
        TS := OpenForReading (file);
        IF (info^.part = HEADER_FIELDS) OR (info^.part = HEADER_FIELDS_NOT) THEN
            SendFragmentedFilePart (RC, TS, info, SD);
        ELSE
            SendContiguousFilePart (RC, TS, info, SD);
        END (*IF*);
        CloseTS (TS);
    END SendFilePart;

(************************************************************************)

PROCEDURE SendOneSection (RC: ReplyCxt;  VAR (*IN*) file: FilenameString;
                                      VAR (*INOUT*) info: BodyRequestInfo;
                                      SD: StructureData);

    (* Sends back the specified part of the file.  This is a response   *)
    (* to a FETCH BODY[<section>]<<partial>>, where <section> is an     *)
    (* optional section number followed optionally by the specification *)
    (* TEXT or the specification of which header fields are wanted.     *)

    CONST
        HeaderSections = BodyParts {HEADER_FIELDS, HEADER_FIELDS_NOT, HEADER, MIME};

    VAR offsetbuff: ARRAY [0..15] OF CHAR;
        label: ARRAY [0..63] OF CHAR;
        label1: ARRAY [0..21] OF CHAR;

    BEGIN
        (* Construct a label to use in the reply.  *)

        Strings.Assign (info^.section, label);
        CASE info^.part OF
             | ALL:     Strings.Assign ("", label1);
             | HEADER,
               HEADER_FIELDS,
               HEADER_FIELDS_NOT:
                        Strings.Assign ("HEADER", label1);
             | MIME:    Strings.Assign ("MIME", label1);
             | TEXT:    Strings.Assign ("TEXT", label1);
             | ELSE
                        Strings.Assign ("", label1);
        END (*CASE*);

        IF (label[0] <> Nul) AND (label1[0] <> Nul) THEN
            Strings.Append ('.', label);
        END (*IF*);
        Strings.Append (label1, label);

        PartialReply (RC, "BODY[");
        PartialReply (RC, label);
        PartialReply (RC, "]");
        IF info^.partial THEN
            PartialReply (RC, "<");
            CardinalToStringLJ (info^.initialoffset, offsetbuff);
            PartialReply (RC, offsetbuff);
            PartialReply (RC, ">");
        END (*IF*);
        SendFilePart (RC, file, info, SD);

    END SendOneSection;

(************************************************************************)

PROCEDURE SendSection (RC: ReplyCxt;  VAR (*IN*) file: FilenameString;
                                      VAR (*INOUT*) info: BodyRequestInfo;
                                      VAR (*INOUT*) SD: StructureData);

    (* Sends back the specified part of the file.  This is a response   *)
    (* to a FETCH BODY[<section>]<<partial>>, where <section> is an     *)
    (* optional section number followed optionally by the specification *)
    (* TEXT or the specification of which header fields are wanted.     *)

    VAR TS: TextStream;

    BEGIN
        IF info <> NIL THEN

            (* Create the structure data if it doesn't already exist. *)

            IF CAST(ADDRESS,SD) = NIL THEN
                TS := OpenForReading (file);
                BuildStructureData (TS, SD);
                CloseTS (TS);
            END (*IF*);

            SendOneSection (RC, file, info, SD);

        END (*IF*);
        DiscardStructureData (SD);

    END SendSection;

(************************************************************************)

END Messages.

